#
#  gpsman --- GPS Manager: a manager for GPS receiver data
#
#  Copyright (c) 2000 Miguel Filgueiras (mig@ncc.up.pt) / Universidade do Porto
#
#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.
#
#  File: map.tcl
#  Last change: 14 October 2000
#
# Includes contributions by Brian Baulch (baulchb@onthenet.com.au)
#  marked "BSB contribution"

## tags used:
#    waypoints: WP WP=$name forWP=$ix lab=$name sq2
#               possibly: inRT=$RTix inRT=:$number
#    labels of WP: WP WPn forWP=$ix lab=$name txt
#               possibly: inRT=$RTix inRT=:$number
#    symbols of WP: WP WPsy lab=$name syforWP=$ix
#    menubuttons assoc to WP: menuWP=$ix lab=$name
#    lines in RTs: RT forRT=$ix from=$itWP to=$itWP line
#    labels of RT stages: RT forRT=$ix label
#    trackpoints: TR forTR=$ix inTR=$ix lab=$ix-$number sq2
#               on first point: TRfirst TR=$ix
#    labels of TP: TR forTR=$ix inTR=$ix lab=$ix-$number txt
#    lines in TRs: TR forTR=$ix line
#    for animation:
#      - points: sq2 an=$no
#        possibly: lastfor=$no
#      - lines: line an=$no
#    when saving map: temp
#    for RT under definition:    mkRT
#      - last line: mkRTlast mkRTend
#      - point under cursor: mkRTlast mkRTcursor
#      - leg: mkRTedge mkRTix=$number from=$itWP to=$itWP line
#    background images: mapimage forIm=$dx,$dy
#    when loading background image ($n in {1, 2}):
#      - WP name to place when geo-referencing map: mapfix mapfixname
#      - name of 3rd WP when adjusting map: mapfix mapfixthird
#      - temporary lines when fixing map: mapfix mapfixline=$n
#      - lines when adjusting map: mapadjust mapfixline=$n
#      - temporary points when adjusting map: mapfix mappoint=$n
##

### set default projection

MapProjectionIs UTM

### bindings

proc SetMapBindings {} {
    # set cursor and initial bindings for map items and perform other
    #  initializations
    # a logo or "dummy" text is created for this purpose and then destroyed
    global Map Logo MAPTYPES MAPW2 MAPH2

    $Map configure -cursor crosshair
    bind $Map <Enter> "focus $Map ; MapCursor"
    bind $Map <Leave> { focus . ; UnMapCursor }
    bind $Map <Motion> { MapCursorMotion %x %y }
    # scrolling in N-S, E-W
    bind $Map <Key-Up> { ScrollMap y scroll 1 units ; MapCursorUpdate }
    bind $Map <Key-space> { ScrollMap y scroll 1 pages ; MapCursorUpdate }
    bind $Map <Key-Down> { ScrollMap y scroll -1 units ; MapCursorUpdate }
    bind $Map <Key-Delete> { ScrollMap y scroll -1 pages ; MapCursorUpdate }
    bind $Map <Key-Left> { ScrollMap x scroll -1 units ; MapCursorUpdate }
    bind $Map <Key-Right> { ScrollMap x scroll 1 units ; MapCursorUpdate }
    # scrolling in NE-SW, NW-SE
    bind $Map <Shift-Up> { ScrollMap y scroll 1 units
       ScrollMap x scroll 1 units ; MapCursorUpdate }
    bind $Map <Shift-Down> { ScrollMap y scroll -1 units
       ScrollMap x scroll -1 units ; MapCursorUpdate }
    bind $Map <Shift-Right> { ScrollMap y scroll 1 units
       ScrollMap x scroll -1 units ; MapCursorUpdate }
    bind $Map <Shift-Left> { ScrollMap y scroll -1 units
       ScrollMap x scroll 1 units ; MapCursorUpdate }
    # panning
    bind $Map <Button-2> "$Map scan mark %x %y"
    bind $Map <B2-Motion> "$Map scan dragto %x %y ; SetVisibleOrigin x ; \
	    SetVisibleOrigin y ; MapCursorUpdate"
    # BSB contribution: wheelmouse scrolling
    bind $Map <Button-5> { ScrollMap y scroll 25 units ; MapCursorUpdate }
    bind $Map <Button-4> { ScrollMap y scroll -25 units ; MapCursorUpdate }
    bind $Map <Shift-Button-5> { ScrollMap y scroll 1 pages
	MapCursorUpdate }
    bind $Map <Shift-Button-4> { ScrollMap y scroll -1 pages
	MapCursorUpdate }
    bind $Map <Control-Button-5> { ScrollMap x scroll 1 pages
	MapCursorUpdate }
    bind $Map <Control-Button-4> { ScrollMap x scroll -1 pages
	MapCursorUpdate }
    bind $Map <Alt-Button-5> { ScrollMap x scroll 25 units
	MapCursorUpdate }
    bind $Map <Alt-Button-4> { ScrollMap x scroll -25 units
	MapCursorUpdate }

    set ts [linsert $MAPTYPES 0 dummy]
    if { [string compare $Logo ""] } {
	$Map create image $MAPW2 $MAPH2 -image $Logo -anchor center -tags $ts
    } else { $Map create text 0 0 -tags $ts }
    foreach m $MAPTYPES {
	$Map bind $m <Enter> { HighLight }
	$Map bind $m <Leave> { LowLight }
    }
    after 5000 "$Map delete dummy"
    bind $Map <Button-1> { MarkMapPoint %x %y }
    # BSB contribution
    bind $Map <Return> { MarkMapPoint %x %y }
    return
}

proc MarkMapPoint {x y} {
    # mark point on map if map is not void
    global Map MapEmpty MapMakingRT MapScale MapLoading MapLoadWPs \
	    MapLoadWPNs MapLoadPos MapLoadWait OVx OVy \
	    CRHAIRx CRHAIRy EdWindow Datum CREATIONDATE COLOUR MESS \
	    DEFAULTSYMBOL DEFAULTDISPOPT

    set xx [expr $OVx+$x-$CRHAIRx] ; set yy [expr $OVy+$y-$CRHAIRy]
    switch -glob $MapLoading {
	0 {
	    if { ! $MapEmpty && ! $MapMakingRT } {
		set os [$Map find overlapping $xx $yy [expr $xx+10] \
			[expr $yy+10]]
		if {! [string compare "" $os] || [OverBackImage $os] } {
		    if { [winfo exists $EdWindow(WP)] } { bell ; return }
		    set f UTM/UPS
		    set p [MapToPosn $xx $yy $f]
		    set opts [list create revert cancel]
		    if { $CREATIONDATE } {
			GMWPoint -1 $opts \
				[FormData WP "PFrmt Posn Datum Date" \
				  [list $f $p $Datum [Now]]]
		    } else {
			GMWPoint -1 $opts \
				[FormData WP "Commt PFrmt Posn Datum" \
				[list [DateCommt [Now]] $f $p $Datum]]
		    }
		}
	    }
	} 
	NoRot=3 {
	    set MapLoadPos(origin,x) $xx ; set MapLoadPos(origin,y) $yy
	    MapCreateWP $xx $yy [lindex $MapLoadWPs 0] [lindex $MapLoadWPNs 0]
	    # get line to 3rd WP and then to 2nd WP
	    foreach a "2 1" {
		set it [$Map find withtag mapfixline=$a]
		$Map dtag $it mapfix ; $Map addtag mapadjust withtag $it
	    }
	    # $it is line to 2nd WP
	    set cs [$Map coords $it]
	    set xl0 [lindex $cs 0] ; set yl0 [lindex $cs 1]
	    set xl1 [lindex $cs 2] ; set yl1 [lindex $cs 2]
	    if { [string compare $MapLoadPos(dir,1) y] } {
		set MapLoadPos(line) [lsort -real "$xl0 $xl1"]
	    } else {
		set MapLoadPos(line) [lsort -real "$yl0 $yl1"]
	    }
	    set MapLoading NoRot=end ; set MapLoadWait 0 ; set MapScale 1e6
	    $Map delete mapfix
	    MapCursor
	}
	NoRot=end {
	    if { $MapLoadPos(scale) > 1e5 } {
		bell
	    } else {
		set MapLoadWait 1
		foreach a "1 2" {
		    set ix [lindex $MapLoadWPs $a]
		    $Map delete forWP=$ix syforWP=$ix
		    eval MapCreateWP $MapLoadPos(adj,$a) $ix \
			    {[lindex $MapLoadWPNs $a]}
		}
		set MapScale $MapLoadPos(scale)
		set MapLoadWait 0
		.wmapload.fr.bns.ok configure -state normal
	    }
	}
	Affine*=[1-3] {
	    # type of transformation and number of points to be placed
	    regexp Affine(.*)=(.*) $MapLoading z tp n
	    incr n -1
	    set MapLoadPos($n,x) $xx ; set MapLoadPos($n,y) $yy
	    MapCreateWP $xx $yy [lindex $MapLoadWPs $n] \
		    [lindex $MapLoadWPNs $n]
	    set MapLoading Affine${tp}=$n
	    $Map delete mapfixname
	    MapCursor
	    if { $n == 0 } {
		.wmapload.fr.bns.ok configure -state normal
	    }
	    # continuation to either MapLoadBkDialDone or MapLoadBkDialCancel
	}
    }
    return
}

proc MapCursor {} {
    # start following pointer on map if map is not void
    global Map MapEmpty MapMakingRT MapRTCurrent MapLoading MapLoadWPNs \
	    MapLoadPos COLOUR

    switch -glob $MapLoading {
	Affine*=[1-3] {
	    $Map delete mapfix
	    regsub .*= $MapLoading "" n
	    incr n -1
	    $Map create text 100 100 -fill $COLOUR(mapsel) -anchor sw \
		    -text [lindex $MapLoadWPNs $n] -justify left \
		    -tags [list map mapfix mapfixname]
	}
	NoRot=3 {
	    $Map delete mapfix
	    foreach a "1 2" {
		set ts [list map mapfix mapfixline=$a]
		eval $Map create line $MapLoadPos(pos,$a) \
			-fill $COLOUR(mapsel) -width 2 -tags {$ts}
	    }
	    $Map create text 100 100 -fill $COLOUR(mapsel) -anchor sw \
		    -text [lindex $MapLoadWPNs 0] -justify left \
		    -tags [list map mapfix mapfixname]
	}
	NoRot=end {
	    $Map delete mapfix
	    # create two circles for 2nd and 3rd WPs
	    foreach a "1 2" {
		$Map create oval 100 100 105 105 \
			-fill $COLOUR(mapsel) -tags [list mapfix mappoint=$a]
	    }
	    $Map create text 100 100 -fill $COLOUR(mapsel) -anchor sw \
		    -text [lindex $MapLoadWPNs 2] -justify center \
		    -tags [list map mapfix mapfixthird]
	    $Map create text 100 100 -fill $COLOUR(mapsel) -anchor sw \
		    -text [lindex $MapLoadWPNs 1] -justify left \
		    -tags [list map mapfix mapfixname]
	}
	0 {
	    if { ! $MapEmpty && $MapMakingRT } {
		set x [lindex $MapRTCurrent 0] ; set y [lindex $MapRTCurrent 1]
		$Map create line $x $y $x $y -fill $COLOUR(mkRT) -arrow first \
			-smooth 0 -width 2 -tags [list mkRT mkRTlast mkRTend]
		$Map create oval [expr $x-3] [expr $y-3] \
			[expr $x+3] [expr $y+3] -fill $COLOUR(mkRT) \
			-tags [list mkRT mkRTlast mkRTcursor]
	    }
	}
    }
    return
}

proc UnMapCursor {} {
    # stop following pointer on map if map is not void
    global Map MapEmpty MapMakingRT MapLoading XCoord YCoord CursorPos

    switch -glob $MapLoading {
	Affine*=* -
	NoRot=* {
	    $Map delete mapfix
	    if { [winfo exists .wmapload] } { Raise .wmapload }
	}
	0 {
	    if { ! $MapEmpty } {
		catch "unset CursorPos"
		set XCoord "" ; set YCoord ""
		if { $MapMakingRT } {
		    $Map delete mkRTlast
		    if { [winfo exists .gmRT] } { Raise .gmRT }
		}
	    }
	}
    }
    return
}

proc MapCursorMotion {x y} {
    # compute coordinates of pointer on map if map is not void
    global Map MapEmpty MapScale MapUTMZone XCoord YCoord OVx OVy CursorPos \
	    MapMakingRT MapRTCurrent MapLoading MapLoadPos MapLoadWait \
	    CRHAIRx CRHAIRy

    set CursorPos [list $x $y]
    set xx [expr $OVx+$x-$CRHAIRx] ; set yy [expr $OVy+$y-$CRHAIRy]
    switch -glob $MapLoading {
	Affine*=[1-3] {
	    # move name of WP to be placed
	    $Map coords mapfixname $xx $yy
	}
	NoRot=3 {
	    # move name of 1st WP and lines to the other two WPs
	    $Map coords mapfixname $xx $yy
	    foreach a "1 2" {
		$Map coords mapfixline=$a $xx $yy \
		    [expr $xx+$MapLoadPos(dx,$a)] [expr $yy+$MapLoadPos(dy,$a)]
	    }
	}
	NoRot=end {
	    if { $MapLoadWait } { return }
	    # move names and positions of 2nd and 3rd WPs
	    # move 2nd on its line; then place 3rd according to scale
	    set min [lindex $MapLoadPos(line) 0]
	    set max [lindex $MapLoadPos(line) 1]
	    if { [string compare $MapLoadPos(dir,1) y] } {
		if { $xx < $min } {
		    set xx $min
		} elseif { $xx > $max } {
		    set xx $max
		}
		set yy [expr $MapLoadPos(origin,y)+($xx-$MapLoadPos(origin,x))\
			*$MapLoadPos(dy,1)*1.0/$MapLoadPos(dx,1)]
	    } else {
		if { $yy < $min } {
		    set yy $min
		} elseif { $yy > $max } {
		    set yy $max
		}
		set xx [expr $MapLoadPos(origin,x)+($yy-$MapLoadPos(origin,y))\
			*$MapLoadPos(dx,1)*1.0/$MapLoadPos(dy,1)]
	    }
	    $Map coords mappoint=1 [expr $xx-3] [expr $yy-3] \
		    [expr $xx+3] [expr $yy+3]
	    $Map coords mapfixname $xx [expr $yy-8]
	    # compute scale (m/pixel)
	    if { [set dpx [expr $xx-$MapLoadPos(origin,x)]] != 0 } {
		set sc [expr $MapLoadPos(dmx,1)*1.0/$dpx]
	    } elseif { [set dpy [expr $yy-$MapLoadPos(origin,y)]] != 0 } {
		set sc [expr $MapLoadPos(dmy,1)*1.0/$dpy]
	    } else {
		set sc 1e6
	    }
	    set x3 [expr $MapLoadPos(origin,x)+$MapLoadPos(dmx,2)/$sc]
	    set y3 [expr $MapLoadPos(origin,y)+$MapLoadPos(dmy,2)/$sc]
	    $Map coords mappoint=2 [expr $x3-3] [expr $y3-3] \
		    [expr $x3+3] [expr $y3+3]
	    $Map coords mapfixthird $x3 [expr $y3-8]
	    MapScaleChange $sc
	    set MapLoadPos(adj,1) [list $xx $yy]
	    set MapLoadPos(adj,2) [list $x3 $y3]
	    set MapLoadPos(scale) $sc
	}
	0 {
	    if { ! $MapEmpty } {
		set p [MapToUTM $xx $yy]
		set XCoord [expr round([lindex $p 2])]
		set YCoord [expr round([lindex $p 3])]
		set MapUTMZone [format %d%s [lindex $p 0] [lindex $p 1]]
		if { $MapMakingRT } {
		    set cx [lindex $MapRTCurrent 0]
		    set cy [lindex $MapRTCurrent 1]
		    $Map coords mkRTend $xx $yy $cx $cy
		    $Map coords mkRTcursor [expr $xx-2] [expr $yy-2] \
			    [expr $xx+2] [expr $yy+2]
		}
	    }
	}
    }
    return
}

proc MapCursorUpdate {} {
    # update cursor coordinates after scrolling
    global CursorPos

    if { ! [catch "set CursorPos"] } {
	eval MapCursorMotion $CursorPos
    }
    return
}

proc ScrollMapTo {x0 y0 x y} {
    # scroll map so that point at ($x0,$y0) is shown at ($x,$y),
    #  pixel coordinates relative to canvas origin
    global Map MapRange

    ScrollMap x moveto [expr [lindex [$Map xview] 0]+($x0-$x)/$MapRange(x)]
    ScrollMap y moveto [expr [lindex [$Map yview] 0]+($y0-$y)/$MapRange(y)]
    return
}

proc ScrollMap {dim args} {
    # scroll map and set corresponding coordinate of origin of visible region
    # $dim in {x, y}, $args suitable to {x,y}view command
    global Map

    eval $Map ${dim}view $args
    SetVisibleOrigin $dim
    return
}

proc SetVisibleOrigin {dim} {
    # set coordinate of origin of visible region
    # $dim in {x, y}
    global Map OV$dim MapRange

    set sc [lindex [$Map ${dim}view] 0]
    set OV$dim [expr $sc*$MapRange($dim)+$MapRange(${dim}0)]
    return
}

proc HighLightWP {ix syit} {
    # highlight WP representation
    global Map COLOUR FRAMEIMAGE

    $Map itemconfigure forWP=$ix -fill $COLOUR(mapsel)
    return
    # use of frame leads to problems when "low-lighting" and when making RT
    if { [string compare $syit ""] } {
	set cs [$Map coords $syit]
	# use of tags here leads to infinite loop...
	set fit [eval $Map create image $cs -anchor center -image $FRAMEIMAGE]
	$Map bind $fit <Leave> "$Map delete $fit ; \
		$Map itemconfigure forWP=$ix -fill $COLOUR(WP)"
	$Map bind $fit <Double-1> "$Map delete $fit ; \
		$Map itemconfigure forWP=$ix -fill $COLOUR(WP) ; \
		OpenItem WP $ix"
	$Map bind $fit <Button-3> "MapWPMenu $ix"
    }
    return
}

proc HighLight {} {
    # highlight mapped item where the pointer is currently on
    global Map COLOUR

    set ts [$Map itemcget [set it [$Map find withtag current]] -tags]
    if { [set i [lsearch $ts {forRT=*}]] != -1 } {
	set t [lindex $ts $i]
	$Map itemconfigure $t -fill $COLOUR(mapsel)
	regsub forRT= $t "" ix
	$Map itemconfigure inRT=$ix -fill $COLOUR(mapsel)
	return
    }
    if { [set i [lsearch $ts {forWP=*}]] != -1 } {
	regsub forWP= [lindex $ts $i] "" ix
	set syit [$Map find withtag syforWP=$ix]
	HighLightWP $ix $syit
	return
    }
    if { [set i [lsearch $ts {syforWP=*}]] != -1 } {
	regsub syforWP= [lindex $ts $i] "" ix
	HighLightWP $ix $it
	return
    }
    if { [set i [lsearch $ts {for??=*}]] != -1 } {
	$Map itemconfigure [lindex $ts $i] -fill $COLOUR(mapsel)
    }
    return
}

proc LowLight {} {
    # finish highlighting a mapped item
    global COLOUR Map

    set ts [$Map itemcget [$Map find withtag current] -tags]
    if { [set i [lsearch $ts {forRT=*}]] != -1 } {
	set t [lindex $ts $i]
	$Map itemconfigure $t -fill $COLOUR(RT)
	regsub forRT= $t "" ix
	$Map itemconfigure inRT=$ix -fill $COLOUR(WP)
	return
    }
    if { [set i [lsearch $ts {forWP=*}]] != -1 } {
	$Map itemconfigure [lindex $ts $i] -fill $COLOUR(WP)
	# $Map delete syframe
	## this avoids an infinite loop; don't ask me why...
	# update idletasks
	return
    }
    if { [set i [lsearch $ts {syforWP=*}]] != -1 } {
	regsub syforWP= [lindex $ts $i] "" ix
	$Map itemconfigure forWP=$i -fill $COLOUR(WP)
	# cannot "$Map delete syframe": infinite loop...
	return
    }
    if { [set i [lsearch $ts {forTR=*}]] != -1 } {
	set t [lindex $ts $i]
	$Map itemconfigure $t -fill $COLOUR(TR)
	regsub forTR= $t "" ix
	$Map itemconfigure inTR=$ix -fill $COLOUR(TP)
    }
    return
}

### map bounds

proc SetMapBounds {} {
    # set map bounds according to mapped items and configure map buttons
    global Map MapBounds MapEmpty MapRange MapMakingRT WConf MAPW2 MAPH2 \
	    MAPWIDTH MAPHEIGHT MapTransfTitle

    set MapBounds [$Map bbox all]
    if { [string compare [$Map find all] ""] } {
	# enlarge bounds so that corners can be scrolled to window centre
	set mbs ""
	foreach i "0 1 2 3" d [list $MAPW2 $MAPH2 $MAPW2 $MAPH2] \
		s "-1 -1 1 1" {
	    lappend mbs [expr $s*$d+[lindex $MapBounds $i]]
	}
	set MapBounds $mbs
	foreach d "x y" i "0 1" ii "2 3" l [list $MAPWIDTH $MAPHEIGHT] {
	    set MapRange($d) \
		    [expr [lindex $MapBounds $ii]-[lindex $MapBounds $i]]
	    set MapRange(${d}0) [lindex $MapBounds $i]
	}
	set st normal
	$Map configure -scrollregion $MapBounds
	SetVisibleOrigin x ; SetVisibleOrigin y
    } else {
	set st disabled
	set MapEmpty 1 ; set MapTransfTitle ""
	foreach b $WConf(mapdatum) { $b configure -state normal }
	set MapRange(x) $MAPWIDTH ; set MapRange(y) $MAPHEIGHT
	set MapRange(x0) 0 ; set MapRange(y0) 0
	$Map configure -scrollregion [list 0 0 $MAPWIDTH $MAPHEIGHT]
	set MapMakingRT 0
    }
    foreach p $WConf(mapstate) t $WConf(mapstatetype) {
	switch $t {
	    menubutton -
	    button {
		$p configure -state $st
	    }
	    menu {
		$p entryconfigure 0 -state $st
		$p entryconfigure 2 -state $st
	    }
	}
    }
    return
}

### datum

proc ChangeMapDatum {datum args} {
    # change map datum
    #  $args is not used but is needed as this is called-back from
    #  menus built by proc FillDatumMenu (geod.tcl)
    # to be called only when map is empty
    global Datum

    set Datum $datum
    return
}

### scale

proc MapScaleToShow {scale} {
    # compute distance an unit to show for map scale in metre/pixel
    global DSCALE MAPSCLENGTH DTUNIT SUBDTUNIT SUBDSCALE

    if { [set v [expr $DSCALE*$MAPSCLENGTH*$scale/1000.0]] < 1 } {
	set u $SUBDTUNIT ; set v [expr 1.0*$v/$SUBDSCALE]
    } else { set u $DTUNIT }
    return "[format %.2f $v] $u"
}

proc MapScaleFromDist {d} {
    # compute scale in metre/pixel from distance shown on map window
    global DSCALE MAPSCLENGTH

    return [expr $d*1000.0/$DSCALE/$MAPSCLENGTH]
}

proc MapScaleChange {value} {
    # show change in map scale
    #  $value is either a scale in metre/pixel when geo-referencing image,
    #   or distance to show on map window
    global MpW MapLoading DTUNIT SUBDTUNIT SUBDSCALE

    if { $MapLoading != 0 } {
	if { $value > 1e5 } {
	    $MpW.frm.frmap3.cv.val configure -text ?
	    update idletasks
	    return
	}
	set txt [MapScaleToShow $value]
    } else {
	if { $value < 1 } {
	    set u $SUBDTUNIT ; set value [expr 1.0*$value/$SUBDSCALE]
	    if { [expr int($value)] != $value } {
		set value [format %.2f $value]
	    }	    
	} else { set u $DTUNIT }
	set txt "$value $u"
    }
    $MpW.frm.frmap3.cv.val configure -text $txt
    update idletasks
    return
}

proc MapScaleSet {d} {
    # apply map scale change
    #  $d is number of distance units represented by $MAPSCLENGTH pixels
    global Map MapScale MAPW2 MAPH2 MapMakingRT MapRTCurrent MapLoading \
	    MapRange OVx OVy MapTransf MapEmpty MESS

    if { $MapLoading != 0 } { return }
    set s [MapScaleFromDist $d]
    if { $s == $MapScale } { return }
    if { ! $MapEmpty && ! [MapNewScale${MapTransf}Transf $s] } {
	GMMessage $MESS(transfcantscale)
	return
    }
    SetCursor . watch
    MapScaleChange $d
    set r [expr $MapScale*1.0/$s]
    set MapScale $s
    # pixel coordinates of centre, relative to canvas origin after scaling
    set xms [expr $r*($OVx+$MAPW2)] ; set yms [expr $r*($OVy+$MAPH2)]
    # scale map items
    foreach item [$Map find withtag sq2] {
	set cs [$Map coords $item]
	set x0 [expr [lindex $cs 0]] ; set y0 [expr [lindex $cs 1]]
	set dx [expr ($r-1)*$x0] ; set dy [expr ($r-1)*$y0]
	set ts [$Map gettags $item]
	if { [set i [lsearch $ts {lab=*}]] != -1 } {
	    set t [lindex $ts $i]
	} else { set t $item }
	$Map move $t $dx $dy
    }
    foreach item [$Map find withtag line] {
	set cs [Apply [$Map coords $item] Multiply $r]
	eval $Map coords $item $cs
    }
    foreach item [$Map find withtag lab] {
	set cs [Apply [$Map coords $item] Multiply $r]
	eval $Map coords $item $cs
    }
    if { $MapMakingRT } {
	set x [expr $r*[lindex $MapRTCurrent 0]]
	set y [expr $r*[lindex $MapRTCurrent 1]]
	set MapRTCurrent [list $x $y [lindex $MapRTCurrent 2]]
    }
    # compute new bounds and origin of visible part
    SetMapBounds
    # scroll old centre (xms,yms) to new centre
    ScrollMapTo $xms $yms [expr $OVx+$MAPW2] [expr $OVy+$MAPH2]
    ResetCursor .
    return
}

### abstract mapping procedures
## conversions between geodetic positions and map coordinates

proc MapFromPosn {latd longd datum} {
    # compute map coordinates from position
    global MapEmpty MapLoading MapScale MapProjPointProc MapProjInitProc \
	    MapTransf MAPW2 MAPH2 WConf MPData MTData

    if { $MapEmpty && ! $MapLoading } {
	set MapEmpty 0
	catch "unset MPData" ; catch "unset MTData"
	foreach b $WConf(mapdatum) { $b configure -state disabled }
	set pt [$MapProjInitProc [list [list $latd $longd $datum]]]
	# default transformation: no rotation
	# default initial location on map: $MAPW2 $MAPH2
	eval MapInitNoRotTransf $MapScale $pt $MAPW2 $MAPH2
    } else {
	set pt [$MapProjPointProc $latd $longd $datum]
    }
    return [eval MapApply${MapTransf}Transf $pt]
}

proc MapToPosn {xm ym pfmt} {
    # compute position in given format from map coordinates
    global MapProjInvertProc MapTransf

    set pt [MapInvert${MapTransf}Transf $xm $ym]
    return [eval $MapProjInvertProc $pt $pfmt]
}

proc MapToUTM {xm ym} {
    # compute UTM position from map coordinates
    global MapProjInvUTMProc MapTransf

    set pt [MapInvert${MapTransf}Transf $xm $ym]
    return [eval $MapProjInvUTMProc $pt]
}

### display

proc MapCreateWP {x y wpix name} {
    # create WP representation on map
    # return rectangle item
    global Map WPCommt WPSymbol WPDispOpt COLOUR SYMBOLIMAGE ICONHEIGHT

    set its [set it [$Map create rectangle [expr $x-1] [expr $y-1] \
	             [expr $x+1] [expr $y+1] -fill $COLOUR(WP) \
		     -outline $COLOUR(WP) \
		     -tags [list WP WP=$name forWP=$wpix lab=$name sq2]]]
    switch [set o $WPDispOpt($wpix)] {
	name -
	s_name {
	    lappend its [$Map create text $x [expr $y-6-$ICONHEIGHT/2.0] \
		    -text $name -fill $COLOUR(WP) -font fixed -justify center \
		    -tags [list WP WPn forWP=$wpix lab=$name txt]]
	}
	comment -
	s_comment {
	    set t $WPCommt($wpix)
	    lappend its [$Map create text $x [expr $y-6-$ICONHEIGHT/2.0] \
		    -text $t -fill $COLOUR(WP) -font fixed -justify center \
		    -tags [list WP WPn forWP=$wpix lab=$name txt]]
	}
    }
    if { [string first s $o] == 0 } {
	lappend its [$Map create image $x $y -anchor center \
		-image $SYMBOLIMAGE($WPSymbol($wpix)) \
		-tags [list WP WPsy syforWP=$wpix lab=$name]]
    }
    foreach m $its {
	$Map bind $m <Double-1> "OpenItem WP $wpix"
	$Map bind $m <Button-3> "MapWPMenu $wpix"
    }
    return $it
}

proc PutMapWP {ix} {
    # map WP with given index
    # return map item for the rectangle
    global Datum WPName WPPosn WPDatum

    set p [MapFromPosn [lindex $WPPosn($ix) 0] [lindex $WPPosn($ix) 1] \
	              $WPDatum($ix)]
    return [MapCreateWP [lindex $p 0] [lindex $p 1] $ix $WPName($ix)]
}

proc PutMapRT {ix} {
    # map RT with given index
    # fails if RT contains a WP either unknown or being edited
    global RTIdNumber RTWPoints RTStages WPDispl EdWindow GMWPIndex \
	    Map COLOUR MESS DataIndex

    SetCursor . watch
    set id $RTIdNumber($ix)
    set its ""
    foreach wp $RTWPoints($ix) {
	set wpix [IndexNamed WP $wp]
	if {! [string compare [set it [$Map find withtag WP=$wp]] ""] } {
	    if { $wpix == -1 } {
		GMMessage "$MESS(cantmapRTunkn) $wp"
		ResetCursor .
		return -1
	    }
	    if { [winfo exists $EdWindow(WP)] && $GMWPIndex==$wpix } {
		GMMessage "$MESS(cantmapRTed): $wp"
		ResetCursor .
		return -1
	    }
	    set it [PutMapWP $wpix]
	    set WPDispl($wpix) 1
	    SetDisplShowWindows WP $wpix select
	}
	lappend its $it
	$Map addtag inRT=$ix withtag forWP=$wpix
    }
    set cs [$Map coords [set it0 [lindex $its 0]]]
    set x0 [lindex $cs 0] ; set y0 [lindex $cs 1]
    set ixlab $DataIndex(RSlabel)
    foreach it [lrange $its 1 end] st $RTStages($ix) {
	if { [string compare $it ""] } {
	    set cs [$Map coords $it]
	    set x [lindex $cs 0] ; set y [lindex $cs 1]
	    set zs [$Map create line $x0 $y0 $x $y -arrow last -smooth 0 \
		    -fill $COLOUR(RT) -width 2 \
		    -tags [list RT forRT=$ix from=$it0 to=$it line]]
	    if { [string compare [set sl [lindex $st $ixlab]] ""] } {
		set xl [expr ($x0+$x)/2] ; set yl [expr ($y0+$y)/2]
		lappend zs [$Map create text $xl $yl \
		    -text $sl -fill $COLOUR(RT) -font fixed -justify center \
		    -tags [list RT forRT=$ix lab]]
	    }
	    foreach l $zs {
		$Map bind $l <Double-1> "OpenItem RT $ix"
		$Map lower $l $it0
	    }
	    set x0 $x ; set y0 $y ; set it0 $it
	}
    }
    ResetCursor .
    return 1
}

proc PutMapTREls {ix tps datum tags} {
    # map TR elements
    #  $ix is index of TR or -1; used for tagging
    #  $tps is list of TR points with given $datum
    #  $tags is tags to add to all created canvas items (may be void)
    global COLOUR Map

    SetCursor . watch
    set tags1 [linsert $tags 0 TR forTR=$ix inTR=$ix]
    set tags2 [linsert $tags 0 TR forTR=$ix line]
    set its "" ; set i 1
    foreach tp $tps {
	set p [MapFromPosn [lindex $tp 0] [lindex $tp 1] $datum]
	set x [lindex $p 0] ; set y [lindex $p 1]
	set it [$Map create rectangle [expr $x-1] [expr $y-1] \
		  [expr $x+1] [expr $y+1] -fill $COLOUR(TP) \
		  -outline $COLOUR(TP) -tags [linsert $tags1 0 lab=$ix-$i sq2]]
	if { $i == 1 } {
	    $Map addtag TRfirst withtag $it
	    $Map addtag TR=$ix withtag $it
	}
	$Map bind $it <Double-1> "OpenItem TR $ix"
	lappend its $it
	if { $i%10 == 0 } {
	    set t [$Map create text $x [expr $y-8] -text $i \
		    -fill $COLOUR(TP) -font fixed -justify center \
		    -tags [linsert $tags1 0 lab=$ix-$i txt]]
	    $Map bind $t <Double-1> "OpenItem TR $ix"
	}
	incr i
    }
    set cs [$Map coords [set it0 [lindex $its 0]]]
    set x0 [lindex $cs 0] ; set y0 [lindex $cs 1]
    foreach it [lrange $its 1 end] {
	if { [string compare $it ""] } {
	    set cs [$Map coords $it]
	    set x [lindex $cs 0] ; set y [lindex $cs 1]
	    set l [$Map create line $x0 $y0 $x $y -smooth 0 \
		    -fill $COLOUR(TR) -width 2 -tags $tags2]
	    $Map bind $l <Double-1> "OpenItem TR $ix"
	    $Map lower $l $it0
	    set x0 $x ; set y0 $y
	}
    }
    ResetCursor .
    return 1
}

proc PutMapTR {ix} {
    # map TR with given index
    global TRTPoints TRDatum

    PutMapTREls $ix $TRTPoints($ix) $TRDatum($ix) ""
    return 1
}

proc PutMapGREl {wh ix} {
    # map GR element of given kind and index
    # fail if the element cannot be unmapped/mapped
    global ${wh}Displ

    if { [set ${wh}Displ($ix)] } {
	if { ! [UnMap $wh $ix] } { return 0 }
    }
    return [PutMap $wh $ix]
}

proc PutMapGR {ix} {
    # map GR with given index
    # may fail if an element cannot be unmapped/mapped
    global GRConts

    set r 1
    foreach p $GRConts($ix) {
	set wh [lindex $p 0]
 	foreach e [lindex $p 1] {
	    if { [set ex [IndexNamed $wh $e]]==-1 || ![PutMapGREl $wh $ex] } {
		set r -1
	    }
 	}
    }
    return $r
}

proc PutMap {wh ix} {
    # put item with index $ix and of type $wh (in $TYPES) on map
    #  if possible
    # set map bounds and change display button in show windows
    global Map ${wh}Displ

    set r [PutMap$wh $ix]
    SetMapBounds
    if { $r == -1 } {
	set [set wh]Displ($ix) 0
	return 0
    }
    set [set wh]Displ($ix) 1
    SetDisplShowWindows $wh $ix select
    return 1
}

proc PutMapAnimPoint {mpos no centre} {
    # display point for animation $no at map position given by
    #  first two elements of $mpos; scroll to centre if $centre
    # draw line from previous point if there is one
    global Map COLOUR OVx OVy MAPW2 MAPH2 FRAMEIMAGE

    set x [lindex $mpos 0] ; set y [lindex $mpos 1]
    if { "[set itl [$Map find withtag lastfor=$no]]" != "" } {
	set cs [$Map coords $itl]
	set x1 [lindex $cs 0] ; set y1 [lindex $cs 1]
	$Map create line $x $y $x1 $y1 -smooth 0 \
		-fill $COLOUR(anim) -width 2 -tags [list an=$no line]
	$Map dtag $itl lastfor=$no
	set blit [$Map find withtag anblink=$no]
	$Map coords $blit $x $y
    } else {
	$Map create image $x $y -anchor center \
		-image $FRAMEIMAGE -tags [list an=$no anblink=$no]
	after 500 "MapBlink anblink=$no 1"
    }
    set it [$Map create rectangle [expr $x-1] [expr $y-1] [expr $x+1] \
	    [expr $y+1] -fill $COLOUR(anim) -outline $COLOUR(anim) \
	    -tags [list an=$no lastfor=$no sq2]]]
    SetMapBounds
    if { $centre } {
	# scroll new point to centre
	ScrollMapTo $x $y [expr $OVx+$MAPW2] [expr $OVy+$MAPH2]
    }
    return
}

proc MapBlink {tag state} {
    # make items with $tag blink on map
    #  $state toggles between 1 and 0
    global Map

    set on 0
    foreach it [$Map find withtag $tag] {
	set on 1
	if { $state } { $Map lower $it } else { $Map raise $it }
    }
    if { $on } { after 500 "MapBlink $tag [expr 1-$state]" }
    return
}

proc UnMapWP {ix} {
    # delete WP with index $ix from map
    # fails if WP belongs to a mapped RT
    global Map WPName

    set it [$Map find withtag WP=$WPName($ix)] ; set ts [$Map gettags $it]
    if { [lsearch $ts {inRT=*}] == -1 } {
	$Map delete forWP=$ix syforWP=$ix menuWP=$ix
	return 1
    }
    return 0
}

proc UnMapRT {ix} {
    # delete RT with index $ix from map
    global Map

    $Map delete forRT=$ix
    foreach it [$Map find withtag inRT=$ix] {
	$Map dtag $it inRT=$ix
    }
    return 1
}

proc UnMapTR {ix} {
    # delete TR with index $ix from map
    global Map

    $Map delete forTR=$ix
    return 1
}

proc UnMapGR {ix} {
    # delete from map all items in GR with index $ix or in its subgroups
    # unmapping of some items may fail, but others will be unmapped
    global GRConts

    set r 1
    set wps ""
    foreach p $GRConts($ix) {
	set wh [lindex $p 0]
	if { [string compare $wh WP] } {
	    foreach e [lindex $p 1] {
		if { [set eix [IndexNamed $wh $e]]==-1 || ![UnMap $wh $eix] } {
		    set r 0
		}
	    }
	} else { set wps [concat $wps [lindex $p 1]] }
    }
    foreach wp $wps {
	if { [set eix [IndexNamed WP $wp]]==-1 || ![UnMap WP $eix] } {
	    set r 0
	}
    }
    return $r
}

proc UnMap {wh ix} {
    # delete item with index $ix and of type $wh (in $TYPES) from map
    # if possible
    global Map ${wh}Displ

    if { [set r [UnMap$wh $ix]] } { 
	set [set wh]Displ($ix) 0
	SetDisplShowWindows $wh $ix deselect
    }
    SetMapBounds
    return $r
}

proc MoveOnMap {wh ix oldname diffname newname} {
    # change mapped item with index $ix
    #  $wh in $TYPES
    #  if $diffname is set $oldname is different from $newname
    global WPDispOpt Map WPName MapMakingRT MapRTCurrent

    if { [string compare $wh WP] } {
	UnMap $wh $ix ; PutMap $wh $ix
    } else {
	# change WP
	set it [$Map find withtag WP=$oldname]
	set ts [$Map gettags $it]
	if { [set iz [lsearch $ts {inRT=*}]] == -1 } {
	    UnMap WP $ix ; PutMap $wh $ix
	    return
	}
	$Map delete forWP=$ix syforWP=$ix
	PutMap WP $ix
	# add  inRT=*  tags
	while { 1 } {
	    set t [lindex $ts $iz]
	    regsub inRT= $t "" rx
	    $Map addtag inRT=$rx withtag forWP=$ix
	    set ts [lrange $ts [expr $iz+1] end]
	    if { [set iz [lsearch $ts {inRT=*}]] == -1 } { break }
	}
	set ni [$Map find withtag WP=$WPName($ix)]
	set x [$Map coords $ni]
	set y [lindex $x 1] ; set x [lindex $x 0]
	if { $MapMakingRT && [lindex $MapRTCurrent 2]==$it } {
	    set MapRTCurrent [list $x $y $ni]
	}
	foreach lf [$Map find withtag from=$it] {
	    $Map dtag $lf from=$it ; $Map addtag from=$ni withtag $lf
	    set cs [lreplace [$Map coords $lf] 0 1 $x $y]
	    eval $Map coords $lf $cs
	}
	foreach lt [$Map find withtag to=$it] {
	    $Map dtag $lt to=$it ; $Map addtag to=$ni withtag $lt
	    set cs [lreplace [$Map coords $lt] 2 3 $x $y]
	    eval $Map coords $lt $cs
	}
    }
    return
}

proc SaveMap {} {
    # save map as a postscript file
    global MpW Map MapBounds PAPERSIZE PAGEWIDTH PAGEHEIGHT GFMode GFRot \
	    GFPaper GFScale GFLegend GFVisible MAPSCLENGTH COLOUR MESS TXT

    set leg "=$TXT(legend)"
    set isc "@$TXT(incscale)" ; set GFScale 1
    set vis "@$TXT(psvisible)" ; set GFVisible 0
    set cms [list $TXT(colour) $TXT(grey) $TXT(mono)]
    set wrs [list $TXT(portr) $TXT(landsc)]
    set pps [array names PAGEWIDTH]
    if { [set i [lsearch -exact $pps $PAPERSIZE]] != 0 } {
	set pps [linsert [lreplace $pps $i $i] 0 $PAPERSIZE]
    }
    set fn [GMGetFileName $TXT(saveto) Plot w \
	            "GFLegend GFScale GFVisible GFMode GFRot GFPaper" \
	            [list $leg $isc $vis $cms $wrs $pps]]
    if { [string compare $fn ".."] } {
	SetCursor . watch
	foreach m $cms tm "color gray mono" {
	    if {! [string compare $GFMode $m] } {
		set GFMode $tm
		break
	    }
	}
	if { $GFVisible } {
	    set mbs $MapBounds
	} else { set mbs [$Map bbox all] }
	set y0 [lindex $mbs 1] ; set x0 [lindex $mbs 0]
	set yf [lindex $mbs 3] ; set xf [lindex $mbs 2]
	set sm [expr ($xf+$x0)/2.0]
	if { $GFScale } {
	    set ys [expr $yf+60]
	    set m $MAPSCLENGTH ; set m2 [expr $m/2.0]
	    scan [$MpW.frm.frmap3.cv.val cget -text] "%d %s" sc su
	    if { [set xs0 [expr $sm-$m2]] < $x0 } {
		set m $m2
		if { [set xs0 [expr $sm-$m/2.0]] < $x0 } {
		    set x0 $xs0
		}
		set sc [expr $sc/2.0]
		if { int($sc) != $sc } {
		    set sc [format "%.1f" $sc]
		} else { set sc [expr int($sc)] }
	    }
	    if { [set xsf [expr $xs0+$m]] > $xf } {
		set xf $xsf
	    }
	    set yd [expr $ys-7.5] ; set yf [expr $ys+7.5]
	    $Map create line $xs0 $ys $xsf $ys -arrow both \
		    -fill $COLOUR(mapsel) -tags temp
	    $Map create line $xs0 $yd $xs0 $yf -fill $COLOUR(mapsel) -tags temp
	    $Map create line $xsf $yd $xsf $yf -fill $COLOUR(mapsel) -tags temp
            $Map create text $sm [expr $ys-8] -font fixed \
		    -fill $COLOUR(mapsel) -text "$sc $su" -tags temp
	}
	if { [string compare $GFLegend ""] } {
	    set yt [expr $yf+20]
	    $Map create text $sm $yt -font fixed \
		    -fill $COLOUR(mapleg) -text $GFLegend -tags temp
	    set yf [expr $yt+5]
	}
	set rot [expr [string compare $GFRot $TXT(portr)]!=0]
	set h [expr $yf-$y0+100] ; set w [expr $xf-$x0+100]
	if { $h > $w } {
	    set sc "-pageheight"
	    if { $rot } {
		set sca $PAGEWIDTH($GFPaper)
	    } else { set sca $PAGEHEIGHT($GFPaper) }
	} else {
	    set sc "-pagewidth"
	    if { $rot } {
		set sca $PAGEHEIGHT($GFPaper)
	    } else { set sca $PAGEWIDTH($GFPaper) }
	}
	if { $GFVisible } {
	    # as in Brian Baulch version
	    $Map postscript -file $fn -colormode $GFMode $sc $sca -rotate $rot
	} else {
	    $Map postscript -file $fn -width $w -height $h -colormode $GFMode \
		    -x [expr $x0-50] -y [expr $y0-50] $sc $sca -rotate $rot
	}
	$Map delete temp
	ResetCursor .
    }
    return
}

proc ClearMap {} {
    # clear map after confirmation
    global MESS

    if { [GMConfirm $MESS(okclrmap)] } {
	DoClearMap
    }
    return
}

proc DoClearMap {} {
    # delete all map items even if being edited
    global MpW Map MapLoading MapScale MapScInitVal MapImageItems \
	    MapImageFile WConf XCoord YCoord MapUTMZone \
	    EdWindow GMWPData GMWPDispl GMRTData GMRTDispl GMTRData \
	    GMGRData GMTRDispl WPDispl RTDispl TRDispl GRDispl TYPES

    # RTs (if they exist) must be dealt with first
    if { [set i [lsearch -exact $TYPES RT]] != -1 } {
	set types [linsert [lreplace $TYPES $i $i] 0 RT]
    } else {
	set types $TYPES
    }
    foreach wh $types {
	if { [winfo exists $EdWindow($wh)] } {
	    set GM${wh}Displ 0
	    set GM${wh}Data [lreplace [set GM${wh}Data] end end 0]
	    $EdWindow($wh).fr.displayed deselect
	}
	foreach n [array names ${wh}Displ] {
	    set ${wh}Displ($n) 0
	}
    }
    eval $Map delete [$Map find all]
    set MapImageItems "" ; catch "unset MapImageFile"
    SetMapBounds
    set MapLoading 0
    set XCoord "" ; set YCoord "" ; set MapUTMZone ""
    $MpW.frm.frmap3.mn configure -state normal
    foreach b $WConf(mapdatum) { $b configure -state normal }
    MapScaleChange $MapScInitVal
    set MapScale [MapScaleFromDist $MapScInitVal]
    StatusMapBackMenu disabled
    return
}

proc MapWPMenu {ix} {
    # create menubutton and menus to put, on map, items in relation to
    #  mapped WP with given index, or for starting making a RT from it
    global Map TXT WPName LsW MAXMENUITEMS MapBounds DSCALE

    destroy $Map.m
    set wp $WPName($ix)
    set mapitem [$Map find withtag WP=$wp]
    menubutton $Map.m -menu $Map.m.mn -text $TXT(withWP)
    set cs [$Map coords $mapitem]
    set sx [lindex $cs 0] ; set sy [lindex $cs 1]
    set y [expr $sy+10] ; set x [expr $sx+50]
    if { $x < [lindex $MapBounds 0] } {
	set x [lindex $MapBounds 0]
    } elseif { $x > [lindex $MapBounds 2] } {
	set x [lindex $MapBounds 2]
    }
    if { $y < [lindex $MapBounds 1] } {
	set y [lindex $MapBounds 1]
    } elseif { $y > [lindex $MapBounds 3] } {
	set y [lindex $MapBounds 3]
    }
    set it [$Map create window $x $y -window $Map.m \
	                -tags [list menuWP=$ix lab=$wp]]
    menu $Map.m.mn -tearoff 0
    $Map.m.mn add command -label $TXT(startRT) \
	    -command "destroy $Map.m ; MapMakeRT $ix $sx $sy"
    $Map.m.mn add cascade -label "$TXT(displ) ..." -menu $Map.m.mn.d
    menu $Map.m.mn.d -tearoff 0
    $Map.m.mn.d add cascade -label "$TXT(within) ..." -menu $Map.m.mn.d.within
    menu $Map.m.mn.d.within -tearoff 0
    foreach d "5 10 20 50 100 200 300 500" {
	$Map.m.mn.d.within add command -label $d \
		-command "destroy $Map.m ; MapWPsWithin [expr $d/$DSCALE] $ix"
    }
    $Map.m.mn.d add cascade -label "$TXT(inrect) ..." -menu $Map.m.mn.d.rect
    set mw $Map.m.mn.d.rect
    menu $mw -tearoff 0
    set n 0 ; set m 0
    foreach it [$LsW.frlWP.frl.box get 0 end] {
	if { [string compare $wp $it] } {
	    if { $n > $MAXMENUITEMS } {
		$mw add cascade -label "$TXT(more) ..." -menu $mw.m$m
		set mw $mw.m$m ; menu $mw -tearoff 0
		set n 0 ; incr m
	    }
	    $mw add command -label $it \
		-command "destroy $Map.m ; MapWPsInRect $ix $it"
	    incr n
	}
    }
    $Map.m.mn.d add cascade -label "$TXT(nameRT) ..." -menu $Map.m.mn.d.rts
    menu $Map.m.mn.d.rts -tearoff 0
    $Map.m.mn.d.rts add command -label $TXT(forthisWP) \
	    -command "destroy $Map.m ; MapRTsFor $ix"
    $Map.m.mn.d.rts add command -label $TXT(formappedWPs) \
	    -command "destroy $Map.m ; MapRTsForMappedWPs"
    $Map.m.mn add command -label $TXT(cancel) -command "destroy $Map.m"
    return
}

proc MapMakeRT {wpix x y} {
    # start making and mapping a RT for a mapped WP
    global Map MapMakingRT MapRTCurrent MapRTLast EdWindow WPName COLOUR TXT

    if { $MapMakingRT } { bell ; return }
    if { [winfo exists $EdWindow(RT)] } { Raise $EdWindow(RT) ; bell ; return }
    set MapMakingRT 1 ; set MapRTLast 0
    set n $WPName($wpix)
    GMRoute -1 {create cancel} [FormData RT "WPoints Displ" [list [list $n] 1]]
    .gmRT.fr.fr3.frbt.del configure -text $TXT(dellast)
    set MapRTCurrent [list $x $y [$Map find withtag WP=$n]]
    $Map addtag inRT=:-1 withtag forWP=$wpix
    $Map create line $x $y $x $y -fill $COLOUR(mkRT) -arrow first -smooth 0 \
	    -width 2 -tags [list mkRT mkRTlast mkRTend]
    $Map create oval [expr $x-3] [expr $y-3] [expr $x+3] [expr $y+3] \
	    -fill $COLOUR(mkRT) -tags [list mkRT mkRTlast mkRTcursor]
    $Map bind mkRTlast <Button-3> { MapFinishRT %x %y }
    $Map bind mkRTlast <Shift-2> { MapCancelRT 1 }
    $Map bind mkRTlast <Shift-1> { MapDelFromRT }
    $Map bind mkRTlast <Button-1> { MapAddToRT 1 %x %y }
    return
}

proc MapFinishRT {x y} {
    global MapMakingRT TXT

    if { $MapMakingRT } {
	MapAddToRT 0 $x $y
	MapDestroyRT
	set frbt .gmRT.fr.fr3.frbt
	$frbt.del configure -text $TXT(del)
	foreach b "insb insa repl inv" {
	    $frbt.$b configure -state normal
	}
	.gmRT.fr.displayed configure -state normal
    }
    return
}

proc MapAddToRT {protest x y} {
    global Map MapMakingRT MapRTCurrent MapRTLast OVx OVy WPName \
	    CRHAIRx CRHAIRy COLOUR

    if { $MapMakingRT } {
	set xx [expr $OVx+$x-$CRHAIRx] ; set yy [expr $OVy+$y-$CRHAIRx]
	set its [$Map find overlapping [expr $xx-3] [expr $yy-3] \
		                       [expr $xx+3] [expr $yy+3]]
	foreach it $its {
	    set ts [$Map gettags $it]
	    if { [set i [lsearch $ts {*forWP=*}]] != -1 } {
		set t [lindex $ts $i]
		regsub .*forWP= $t "" ix
		$Map addtag inRT=:$MapRTLast withtag forWP=$ix
		set n $WPName($ix)
		set toit [$Map find withtag WP=$n]
		set cs [$Map coords $toit]
		set xx [lindex $cs 0] ; set yy [lindex $cs 1]
		$Map coords mkRTlast $xx $yy $xx $yy
		$Map create line $xx $yy \
			[lindex $MapRTCurrent 0] [lindex $MapRTCurrent 1] \
			-fill $COLOUR(mkRT) -arrow first -smooth 0 -width 2 \
			-tags [list mkRT mkRTedge mkRTix=$MapRTLast \
			       to=[lindex $MapRTCurrent 2] from=$toit line]
		set MapRTCurrent [list $xx $yy $toit]
		GMRTChange insa $n
		incr MapRTLast
		.gmRT.fr.fr3.frbt.del configure -state normal
		return
	    }
	}
	if { $protest } { bell }
    }
    return
}

proc MapDelFromRT {} {
    # delete last WP from RT being built on map but fail if there is
    #  only one
    # GMRTChange will call MapDelRTLast on success
    global MapMakingRT MapRTLast

    if { $MapMakingRT } {
	if { $MapRTLast == 0 } { bell ; return }
	GMRTChange del
    }
    return
}

proc MapDelRTLast {prevwp} {
    # update map by deleting last WP on RT under construction on map
    global Map MapRTLast MapRTCurrent

    incr MapRTLast -1
    set it [$Map find withtag WP=$prevwp]
    set cs [$Map coords $it]
    set xx [lindex $cs 0] ; set yy [lindex $cs 1]
    $Map coords mkRTlast $xx $yy $xx $yy
    set MapRTCurrent [list $xx $yy $it]
    # zero or one items will have this tag
    foreach it [$Map find withtag mkRTix=$MapRTLast] {
	$Map delete $it
    }
    foreach it [$Map find withtag inRT=:$MapRTLast] {
	$Map dtag $it inRT=:$MapRTLast
    }
    return
}

proc MapCancelRT {fromMap} {
    # cancel construction of RT on map
    global MapMakingRT

    if { $MapMakingRT } {
	MapDestroyRT
	if { $fromMap } { GMButton RT cancel }
    }
    return
}

proc MapDestroyRT {} {
    # destroy RT being made on map
    global Map MapMakingRT MapRTLast

    set MapMakingRT 0
    $Map delete mkRT
    while { $MapRTLast >= 0 } {
	incr MapRTLast -1
	foreach it [$Map find withtag inRT=:$MapRTLast] {
	    $Map dtag $it inRT=:$MapRTLast
	}
    }
    return
}

proc MapWPsWithin {d ix} {
    # map all WPs with distance $d of WP with index $ix
    global WPName WPPosn WPDatum WPDispl EdWindow GMWPIndex GMWPDispl

    SetCursor . watch
    if { [winfo exists $EdWindow(WP)] } {
	set edix $GMWPIndex
    } else { set edix -1 }
    set p1 $WPPosn($ix) ; set d1 $WPDatum($ix)
    SetDatumData $d1
    foreach ix2 [array names WPName] {
	# current wp is displayed so there is no need to check $ix!=$ix2
	if { ! $WPDispl($ix2) } {
	    set p2 $WPPosn($ix2) ; set d2 $WPDatum($ix2)
	    if  { [string compare $d1 $d2] } {
		set p2 [ConvertDatum [lindex $p2 0] [lindex $p2 1] $d2 $d1 DDD]
	    }
	    if { $d >= [lindex [ComputeDistFD $p1 $p2] 0] } {
		if { $ix2 == $edix } {
		    if { ! $GMWPDispl } {
			PutMapWP $ix2
			set GMWPDispl 1
			$EdWindow(WP).fr.displayed select
		    }
		} else { PutMapWP $ix2 }
		set WPDispl($ix2) 1
	    }
	}
    }
    SetMapBounds
    ResetCursor .
    return
}

proc MapWPsInRect {ix1 wp2} {
    # map all WPs in the rectangle defined by the WPs with index $ix1
    #  and name $wp2
    global WPName WPPosn WPDatum WPDispl EdWindow GMWPIndex GMWPDispl

    SetCursor . watch
    if { [winfo exists $EdWindow(WP)] } {
	set edix $GMWPIndex
    } else { set edix -1 }
    set p1 $WPPosn($ix1) ; set d1 $WPDatum($ix1)
    SetDatumData $d1
    set ix2 [IndexNamed WP $wp2]
    set p2 $WPPosn($ix2) ; set d2 $WPDatum($ix2)
    if  { [string compare $d1 $d2] } {
	set p2 [ConvertDatum [lindex $p2 0] [lindex $p2 1] $d2 $d1 DDD]
    }
    set la1 [lindex $p1 0] ; set lo1 [lindex $p1 1]
    set la2 [lindex $p2 0] ; set lo2 [lindex $p2 1]
    if { $la1 >= $la2 } {
	set lamx $la1 ; set lamn $la2
    } else { set lamx $la2 ; set lamn $la1 }
    if { $lo1 >= $lo2 } {
	set lomx $lo1 ; set lomn $lo2
    } else { set lomx $lo2 ; set lomn $lo1 }
    # first wp assumed to be displayed
    foreach ixn [array names WPName] {
	if { ! $WPDispl($ixn) } {
	    set pn $WPPosn($ixn) ; set dn $WPDatum($ixn)
	    if  { [string compare $d1 $dn] } {
		set pn [ConvertDatum [lindex $pn 0] [lindex $pn 1] $dn $d1 DDD]
	    }
	    set lan [lindex $pn 0]
	    if { $lamx>=$lan && $lan>=$lamn } {
		set lon [lindex $pn 1]
		if { $lomx>=$lon && $lon>=$lomn } {
		    if { $ixn == $edix } {
			if { ! $GMWPDispl } {
			    PutMapWP $ixn
			    set GMWPDispl 1
			    $EdWindow(WP).fr.displayed select
			}
		    } else { PutMapWP $ixn }
		    set WPDispl($ixn) 1
		}
	    }
	}
    }
    SetMapBounds
    ResetCursor .
    return
}

proc MapRTsFor {ix} {
    # map all RTs that contain the WP with index $ix
    global WPRoute RTDispl

    foreach rt $WPRoute($ix) {
	set ixrt [IndexNamed RT $rt]
	if { ! $RTDispl($ixrt) } {
	    if { [PutMapRT $ixrt] } { set RTDispl($ixrt) 1 }
	}
    }
    return
}

proc MapRTsForMappedWPs {} {
    # map all RTs for all mapped WPs
    global WPName WPDispl

    foreach ix [array names WPName] {
	if { $WPDispl($ix) } {
	    MapRTsFor $ix
	}
    }
    return
}

### background image

proc LoadMapBack {} {
    # load map background: either an image to be fixed, or information on
    #  an image and fixing information
    global Map File MESS

    if { [string compare [$Map find all] ""] && \
	    ! [GMConfirm $MESS(clrcurrmap)] } {
	return
    }
    set r [LoadMapFixedBk]
    switch -- [lindex $r 0] {
	0 {
	    LoadMapBackImage $File(MapBkInfo)
	}
	1 {
	    eval LoadMapBackGeoRef [lrange $r 1 end]
	}
	2 {
	    # this is here only for supporting the old (<4.0) file format
	    eval LoadMapBackFixed [lrange $r 1 end]
	}
    }
    return
}

proc BadImage {im filename} {
    # create image

    SetCursor . watch
    catch "image delete $im"
    set r [catch "image create photo $im -file $filename"]
    ResetCursor .
    return $r
}

proc MapCreateOriginImage {path} {
    # create map background image at origin
    # clear the map, disable scale and datum, and set image parameters
    global MpW Map MapImageFile MapImageHeight MapImageWidth MapImageItems \
	    MapImageGrid MAPW2 MAPH2 WConf

    DoClearMap
    $MpW.frm.frmap3.mn configure -state disabled
    $MpW.frm.frmap3.cv.val configure -text ?
    foreach b $WConf(mapdatum) { $b configure -state disabled }
    set MapImageFile(0,0) $path
    set MapImageHeight [image height MapImage]
    set MapImageWidth [image width MapImage]
    set MapImageItems [$Map create image 0 0 -image MapImage \
	    -anchor nw -tags [list map mapimage forIm=0,0]]
    SetMapBounds
    # scroll image to centre it
    ScrollMapTo [expr $MapImageWidth/2.0] [expr $MapImageHeight/2.0] \
	        $MAPW2 $MAPH2
    set MapImageGrid(dxmin) -1 ; set MapImageGrid(dymin) -1
    set MapImageGrid(dxn) 3 ; set MapImageGrid(dyn) 3
    return
}

proc LoadMapBackGeoRef {path datum pdata tdata scale ixps} {
    # load geo-referenced map background image
    #  $pdata, $tdata describe projection and transformation and are pairs
    #   with name and list of pairs with parameter name and value
    #  $ixps: list of image grid coordinates and path for subsidiary images
    global MpW Map MapImageFile MapImageItems MPData MTData MapScale \
	    MapImageHeight MapImageWidth MapImageGrid MapProjection \
	    MapProjTitle MapProjInitProc MapTransf MapEmpty DTUNIT MESS TXT \
	    MAPPROJDATA MAPPARTPDATA MAPPARTPROJ

    foreach ixp $ixps {
	set p [lindex $ixp 1]
	if { [BadImage MapImage[lindex $ixp 0] $p] } {
	    GMMessage "$MESS(badimage): $p"
	    return
	}
    }
    if { [BadImage MapImage $path] } {
	GMMessage "$MESS(badimage): $path"
	return
    }
    MapCreateOriginImage $path
    set dxmin 0 ; set dxmax 0 ; set dymin 0 ; set dymax 0
    foreach ixp $ixps {
	scan $ixp %d,%d%s dx dy p
	set MapImageFile($dx,$dy) $p
	set x [expr $MapImageWidth*$dx] ; set y [expr $MapImageHeight*$dy]
	set it [$Map create image $x $y \
		-image "MapImage$dx,$dy" -anchor nw \
		-tags [list map mapimage forIm=$dx,$dy]]
	$Map lower $it
	lappend MapImageItems $it
	if { $dx > $dxmax } { set dxmax $dx }
	if { $dy > $dymax } { set dymax $dy }
	if { $dx < $dxmin } { set dxmin $dx }
	if { $dy < $dymin } { set dymin $dy }
    }
    set MapImageGrid(dxmin) [expr $dxmin-1]
    set MapImageGrid(dymin) [expr $dymin-1]
    set MapImageGrid(dxn) [expr $dxmax+3-$dxmin]
    set MapImageGrid(dyn) [expr $dymax+3-$dymin]
    SetMapBounds
    ChangeMapDatum $datum
    set MapScale $scale
    catch "unset MPData" ; catch "unset MTData"
    set MapProjection [lindex $pdata 0]
    if { [catch "set mp $MAPPARTPROJ($MapProjection)"] } {
	set MapProjTitle $TXT(PRJ$MapProjection)
	foreach p [lindex $pdata 1] {
	    set MPData([lindex $p 0]) [lindex $p 1]
	}
    } else {
	MapProjectionIs $MapProjection
	foreach e $MAPPROJDATA($mp) v $MAPPARTPDATA($MapProjection) {
	    set MPData($e) $v
	}
    }
    MapTransfIs [lindex $tdata 0]
    foreach p [lindex $tdata 1] {
	set MTData([lindex $p 0]) [lindex $p 1]
    }
    set MapEmpty 0
    set txt [MapScaleToShow $scale]
    $MpW.frm.frmap3.cv.val configure -text $txt
    StatusMapBackMenu normal
    return
}

proc LoadMapBackFixed {path datum origin scale ixps} {
    # load map background image with fixing information in old (<4.0) format
    #  $ixps: list of image grid coordinates and path for subsidiary images
    global MpW Map MapImageFile MapImageItems MapProjInitProc MPData MTData \
	    MapScale MapImageHeight MapImageWidth MapImageGrid MapProjection \
	    MapEmpty DTUNIT MESS TXT

    if { [string compare $MapProjection UTM] && \
	    ! [GMConfirm "$MESS(notUTMproj): $TXT(PRJ$MapProjection)"] } {
	return
    }
    foreach ixp $ixps {
	set p [lindex $ixp 1]
	if { [BadImage MapImage[lindex $ixp 0] $p] } {
	    GMMessage "$MESS(badimage): $p"
	    return
	}
    }
    if { [BadImage MapImage $path] } {
	GMMessage "$MESS(badimage): $path"
	return
    }
    MapCreateOriginImage $path
    set dxmin 0 ; set dxmax 0 ; set dymin 0 ; set dymax 0
    foreach ixp $ixps {
	scan $ixp %d,%d%s dx dy p
	set MapImageFile($dx,$dy) $p
	set x [expr $MapImageWidth*$dx] ; set y [expr $MapImageHeight*$dy]
	set it [$Map create image $x $y \
		-image "MapImage$dx,$dy" -anchor nw \
		-tags [list map mapimage forIm=$dx,$dy]]
	$Map lower $it
	lappend MapImageItems $it
	if { $dx > $dxmax } { set dxmax $dx }
	if { $dy > $dymax } { set dymax $dy }
	if { $dx < $dxmin } { set dxmin $dx }
	if { $dy < $dymin } { set dymin $dy }
    }
    set MapImageGrid(dxmin) [expr $dxmin-1]
    set MapImageGrid(dymin) [expr $dymin-1]
    set MapImageGrid(dxn) [expr $dxmax+3-$dxmin]
    set MapImageGrid(dyn) [expr $dymax+3-$dymin]
    SetMapBounds
    ChangeMapDatum $datum
    set MapScale $scale
    catch "unset MPData" ; catch "unset MTData"
    set latd [lindex $origin 0] ; set longd [lindex $origin 1]
    set pt [$MapProjInitProc [list [list $latd $longd $datum]]]
    eval MapInitNoRotTransf $MapScale $pt [lindex $origin 2] [lindex $origin 3]
    set MapEmpty 0
    set txt [MapScaleToShow $scale]
    $MpW.frm.frmap3.cv.val configure -text $txt
    StatusMapBackMenu normal
    return
}

proc LoadMapBackImage {filename} {
    # load map background image to be geo-referenced from file under $filename
    global MpW Map MapLoading MapLdOldScale MapImageItems MapImageFile \
	    MapImageHeight MapImageWidth MapScInitVal EdWindow Number \
	    MAPKNOWNTRANSFS MESS TXT TYPES

    set ts "" ; set rs ""
    foreach t $MAPKNOWNTRANSFS {
	lappend ts $TXT(TRNSF$t) ; lappend rs $t
    }
    lappend ts $TXT(cancel) ; lappend rs 0
    switch [set how [GMSelect $MESS(georefhow) $ts $rs]] {
	0 {
	    return
	}
	AffineConf {
	    set n 2
	}
	Affine -
	NoRot {
	    set n 3
	}
	default {
	    GMMessage "Bug: missing case for transformation"
	    return
	}
    }
    if { $Number(WP) < $n } {
	GMMessage [format $MESS(cantfix) $n]
	return
    }
    if { [BadImage MapImage $filename] } {
	GMMessage $MESS(badimage)
	return
    }
    MapCreateOriginImage [file join [pwd] $filename]
    if { [scan [$MpW.frm.frmap3.cv.val cget -text] %d MapLdOldScale] != 1 } {
	set MapLdOldScale $MapScInitVal
    }
    foreach wh $TYPES {
	if { [winfo exists $EdWindow($wh)] } {
	    $EdWindow($wh).fr.displayed configure -state disabled
	}
    }
    set MapLoading ${how}=$n
    MapLoadBkDial $how $n
    return
}

proc MapLoadBkDial {how n} {
    # dialog used during map background loading
    #  $how in $MAPKNOWNTRANSFS
    #  $n is number of WPs used for geo-referencing
    global WPName MapLoadWPs MapLoadWPNs NAMEWIDTH TXT MESS COLOUR EPOSX EPOSY

    destroy .wmapload
    set w .wmapload
    toplevel $w
    wm protocol $w WM_DELETE_WINDOW { MapLoadBkCancel }
    wm title $w "GPS Manager: $TXT(mapload)"
    wm transient $w
    wm geometry $w +[expr $EPOSX+100]+$EPOSY

    frame $w.fr -relief flat -borderwidth 5 -bg $COLOUR(messbg)
    label $w.fr.title -text $TXT(mapload) -relief sunken
    message $w.fr.text -aspect 800 -text [format $MESS(mapaskWPs) $n]
    listbox $w.fr.bx -height 3 -width $NAMEWIDTH -relief flat \
 	    -selectmode single -exportselection true
    frame $w.fr.bns -relief flat -borderwidth 0
    button $w.fr.bns.ok -text Ok -command MapLoadBkDialDone -state disabled
    button $w.fr.bns.cnc -text $TXT(cancel) -command MapLoadBkCancel
    pack $w.fr -side top
    pack $w.fr.bns.ok $w.fr.bns.cnc -side left
    pack $w.fr.title $w.fr.text $w.fr.bx $w.fr.bns -side top -pady 5
    update idletasks

    set no $n ; set MapLoadWPs ""
    while { $no > 0 } {
	if {! [string compare "" [set wps [ChooseItems WP]]] } {
	    MapLoadBkCancel
	    return
	}
	foreach ix $wps {
	    set nn $WPName($ix) ; set d 0
	    foreach n [$w.fr.bx get 0 end] {
		if {! [string compare $n $nn] } {
		    set d 1
		    break
		}
	    }
	    if { $d } {
		GMMessage [format $MESS(duplicate) $nn]
	    } else {
		$w.fr.bx insert end $nn
		lappend MapLoadWPs $ix
		incr no -1
		if { $no == 0 } { break }
	    }
	}
    }
    $w.fr.text configure -text $MESS(mapadjust)
    switch $how {
	NoRot {
	    set MapLoadWPNs [$w.fr.bx get 0 end]
	    MapComputePositions
	}
	AffineConf {
	    set MapLoadWPNs [list [$w.fr.bx get 1 1] [$w.fr.bx get 0 0]]
	    set MapLoadWPs [list [lindex $MapLoadWPs 1] [lindex $MapLoadWPs 0]]
	}
	Affine {
	    set MapLoadWPNs [list [$w.fr.bx get 2 2] [$w.fr.bx get 1 1] \
		    [$w.fr.bx get 0 0]]
	    set MapLoadWPs [list [lindex $MapLoadWPs 2] \
		    [lindex $MapLoadWPs 1] [lindex $MapLoadWPs 0]]
	}
    }
    # control will be assumed by MapCursor, MarkMapPoint and MapLoadBkDialDone

    return
}

proc MapLoadRestore {} {
    # restore interface state after success or failure of map loading
    global MapLoading MapLoadPos EdWindow TYPES

    foreach wh $TYPES {
	if { [winfo exists $EdWindow($wh)] } {
	    $EdWindow($wh).fr.displayed configure -state normal
	}
    }
    set MapLoading 0
    destroy .wmapload
    catch "unset MapLoadPos"
    return
}

proc MapLoadBkDialDone {} {
    # successful end of map background loading dialog
    global Map MapLoading MapScale MapLoadWPs MapLoadPos MapEmpty EdWindow \
	    WPDispl GMWPIndex GMWPDispl GMWPData MESS MPData MTData

    catch "unset MTData"
    switch -glob $MapLoading {
	Affine=* {
	    if {! [MapInitAffineTransf] } {
		GMMessage $MESS(cantsolve)
		MapLoadBkCancel
		return
	    }
	}
	AffineConf=* {
	    if {! [MapInitAffineConfTransf] } {
		GMMessage $MESS(cantsolve)
		MapLoadBkCancel
		return
	    }
	}
	NoRot=* {
	    MapInitNoRotTransf $MapScale $MapLoadPos(xt0) $MapLoadPos(yt0) \
		    $MapLoadPos(origin,x) $MapLoadPos(origin,y)
	}
    }
    MapScaleChange $MapScale
    MapLoadRestore
    set MapEmpty 0
    $Map delete mapadjust
    StatusMapBackMenu normal
    set ned ""
    foreach wpix $MapLoadWPs {
	set WPDispl($wpix) 1
	if { [winfo exists $EdWindow(WP)] && $GMWPIndex == $wpix } {
	    set GMWPDispl 1
	    set GMWPData [lreplace $GMWPData end end 1]
	    $EdWindow(WP).fr.displayed select
	} else {
	    lappend ned $wpix
	}
    }
    SetDisplShowWindows WP $ned select
    return
}

proc MapLoadBkCancel {} {
    # cancel loading a map background image
    global MpW Map MapLoading MapLdOldScale MapImageItems

    eval $Map delete [$Map find all]
    set MapImageItems ""
    SetMapBounds
    if { $MapLoading != 0 } {
	MapLoadRestore
	# now $MapLoading is 0
	$MpW.frm.frmap3.mn configure -state normal
	MapScaleChange $MapLdOldScale
    }
    return
}

proc MapGeoRefPoints {n} {
    # compute planar Cartesian coordinates of $n WPs for geo-referencing
    #  and initialize projection procedure
    #  $MapLoadWPs is list of indices of relevant WPs
    # set MapLoadPos(latd), MapLoadPos(longd) to coords of 1st WP, relative
    #  to $Datum
    global Datum MapLoadWPs MapLoadPos WPPosn WPDatum MapProjInitProc \
	    MapProjPointProc MPData

   for { set i 0 ; set ps "" } { $i < $n } { incr i } {
       set ix [lindex $MapLoadWPs $i]
       set p $WPPosn($ix)
       set latd [lindex $p 0] ; set longd [lindex $p 1]
       if { [string compare [set datum $WPDatum($ix)] $Datum] } {
	   set p [ConvertDatum $latd $longd $datum $Datum DDD]
	   set latd [lindex $p 0] ; set longd [lindex $p 1]
       }
       if { $i == 0 } {
	   set MapLoadPos(latd) $latd ; set MapLoadPos(longd) $longd
       }
       lappend ps [list $latd $longd $Datum]
   }
   catch "unset MPData"
   $MapProjInitProc $ps
   set xys ""
   foreach p $ps {
       lappend xys [eval $MapProjPointProc $p]
   }
   return $xys
}

proc MapComputePositions {} {
    # compute lines from 1st to 2nd and 1st to 3rd selected WPs
    #  when loading a map background image with no rotation
    # set MapLoadPos(xt0),MapLoadPos(yt0) to terrain coords of 1st WP
    global MapLoadPos MAPWIDTH MAPHEIGHT

    set tcs [MapGeoRefPoints 3]
    set p0 [lindex $tcs 0]
    set MapLoadPos(xt0) [set xt0 [lindex $p0 0]]
    set MapLoadPos(yt0) [set yt0 [lindex $p0 1]]
    set mx $MAPWIDTH
    if { $MAPHEIGHT > $MAPWIDTH } { set mx $MAPHEIGHT }
    incr mx 10000
    foreach a "1 2" {
	set p [lindex $tcs $a]
	set xta [lindex $p 0] ; set yta [lindex $p 1]
	set MapLoadPos(dmx,$a) [set dx [expr $xta-$xt0]]
	set MapLoadPos(dmy,$a) [set dy [expr $yt0-$yta]]
	if { abs($dx) >= abs($dy) } {
	    set MapLoadPos(dir,$a) x
	} else {
	    set MapLoadPos(dir,$a) y
	}
	set l [expr sqrt(1.0*$dx*$dx+1.0*$dy*$dy)]
        set x [expr 100+$mx/$l*$dx] ; set y [expr 100+$mx/$l*$dy]
	set MapLoadPos(pos,$a) [list 100 100 $x $y]
	set MapLoadPos(dx,$a) [expr $x-100]
	set MapLoadPos(dy,$a) [expr $y-100]
    }
    return
}

proc ClearMapBack {} {
    # clear map background images
    global MpW Map MapImageItems MapImageFile MESS

    if { [GMConfirm $MESS(okclrbkmap)] } {
	$Map delete mapimage
	set MapImageItems "" ; catch "unset MapImageFile"
	$MpW.frm.frmap3.mn configure -state normal
	StatusMapBackMenu disabled
	return 1
    }
    return 0
}

proc OverBackImage {its} {
    # check if list of map items $its contains only a background image item
    global MapImageItems

    if { [llength $its] > 1 } { return 0 }
    foreach it $MapImageItems {
	if { $it == $its } { return 1 }
    }
    return 0
}

proc SaveMapBack {} {
    # save map background image information
    global Map MapImageFile MapScale MapProjection MPData MapTransf MTData \
	    MAPPARTPROJ MAPPROJDATA MAPTRANSFDATA

    if {! [string compare [$Map find withtag mapimage] ""] } { return }
    set pd $MapProjection
    if { [catch "set MAPPARTPROJ($MapProjection)"] } {
	foreach e $MAPPROJDATA($MapProjection) {
	    lappend pd "$e=$MPData($e)"
	}
    }
    set pt $MapTransf
    foreach e $MAPTRANSFDATA($MapTransf) {
	lappend pt "$e=$MTData($e)"
    }
    set l ""
    foreach ixs [array names MapImageFile] {
	if { [string compare $ixs 0,0] } {
	    set l [linsert $l 0 [list $ixs $MapImageFile($ixs)]]
	}
    }
    SaveFile mapback MapBkInfo $MapImageFile(0,0) $pd $pt $MapScale $l
    return
}

proc ChangeMapBack {} {
    # dialog used to change map background images
    global MapImageGrid MapImageFile TXT EPOSX EPOSY COLOUR

    # name .wchgmapbak used explicitly below
    set w .wchgmapbak
    if { [winfo exists $w] } { Raise $w ; return }

    toplevel $w
    wm protocol $w WM_DELETE_WINDOW "destroy $w"
    wm title $w "GPS Manager: $TXT(mpbkchg)"
    wm transient $w
    wm geometry $w +[expr $EPOSX+100]+$EPOSY

    frame $w.fr -relief flat -borderwidth 5 -bg $COLOUR(messbg)
    label $w.fr.title -text $TXT(mpbkchg) -relief sunken
    set rw 43 ; set rh 21
    set wd [expr 3*$rw] ; set hg [expr 3*$rh]
    canvas $w.fr.grid -width $wd -height $hg -relief sunken \
	    -xscrollincrement $rw -yscrollincrement $rh
    set cv $w.fr.grid
    # make central 3x3 grid: canvas 0,0 is upper left corner of -1,-1 slot
    for { set dx -1 } { $dx < 2 } { incr dx } {
	MapColumnBackGrid $cv $dx -1 3 $rw $rh
    }
    # extend grid if needs be
    foreach d "x y" h "Column Row" \
	    omin "-1 $MapImageGrid(dxmin)" on "3 $MapImageGrid(dxn)" {
	if { [set d0 $MapImageGrid(d${d}min)] < -1 } {
	    for { set dd $d0 } { $dd < -1 } { incr dd } {
		Map${h}BackGrid $cv $dd $omin $on $rw $rh
		set bb [$cv bbox all]
		set x0 [lindex $bb 0] ; set y0 [lindex $bb 1]
		set x1 [lindex $bb 2] ; set y1 [lindex $bb 3]
		$cv configure -width [expr $x1-$x0] -height [expr $y1-$y0]
		$cv ${d}view scroll -1 units
	    }
	}
	if { [set df [expr $d0+$MapImageGrid(d${d}n)]] > 1 } {
	    for { set dd 2 } { $dd < $df } { incr dd } {
		Map${h}BackGrid $cv $dd $omin $on $rw $rh
		set bb [$cv bbox all]
		set x0 [lindex $bb 0] ; set y0 [lindex $bb 1]
		set x1 [lindex $bb 2] ; set y1 [lindex $bb 3]
		$cv configure -width [expr $x1-$x0] -height [expr $y1-$y0]
	    }
	}
    }

    foreach ixs [array names MapImageFile] {
	$cv itemconfigure forIm=$ixs -fill $COLOUR(fullgrid)
    }
    $cv itemconfigure forIm=0,0 -width 4 -outline $COLOUR(mapsel)
    set it [$cv create text [expr 1.5*$rw] [expr 1.5*$rh] -anchor center \
	     -text + -justify center]
    $cv bind $it <Enter> "MapBackGridEnter 0 0"
    $cv bind $it <Leave> "MapBackGridLeave 0 0"
    $cv bind $it <Button-1> "MapBackGridSelect 0 0"

    frame $w.fr.cs -relief flat -borderwidth 0
    label $w.fr.cs.tit -text $TXT(mpbkgrcs):
    label $w.fr.cs.cs -text 0,0

    frame $w.fr.pt -relief flat -borderwidth 0
    label $w.fr.pt.tit -text $TXT(file):
    label $w.fr.pt.pt -text $MapImageFile(0,0) -width 50

    frame $w.fr.bns -relief flat -borderwidth 0
    button $w.fr.bns.ld -text $TXT(load) \
	    -command "MapBackGridLoad $rw $rh ; \
	              $w.fr.bns.ld configure -state normal"
    button $w.fr.bns.clr -text $TXT(clear) -state disabled \
	    -command "MapBackGridClear ; $w.fr.bns.clr configure -state normal"
    button $w.fr.bns.clrall -text $TXT(clearall) \
	    -command {
	      if { [ClearMapBack] } {
		  destroy .wchgmapbak
	      } else {
		  .wchgmapbak.fr.bns.clrall configure -state normal
	      }
	    }
    button $w.fr.bns.ok -text Ok -command "destroy $w"

    pack $w.fr -side top
    pack $w.fr.cs.tit $w.fr.cs.cs -side left
    pack $w.fr.pt.tit $w.fr.pt.pt -side left
    pack $w.fr.bns.ld $w.fr.bns.clr $w.fr.bns.clrall $w.fr.bns.ok -side left
    pack $w.fr.title $w.fr.grid $w.fr.cs $w.fr.pt $w.fr.bns -side top -pady 5
    update idletasks

    return
}

proc MapBackGridEnter {dx dy} {
    # cursor on image grid slot
    global COLOUR

    set cv .wchgmapbak.fr.grid
    $cv itemconfigure forIm=$dx,$dy -fill $COLOUR(mapsel)
    return
}

proc MapBackGridLeave {dx dy} {
    # cursor out of image grid slot
    global COLOUR MapImageFile

    set cv .wchgmapbak.fr.grid
    if { [catch "set MapImageFile($dx,$dy)"] } {
	set c emptygrid
    } else { set c fullgrid }
    $cv itemconfigure forIm=$dx,$dy -fill $COLOUR($c)
    return
}

proc MapBackGridSelect {dx dy} {
    # click on an image grid slot
    global COLOUR MapImageFile

    set fr .wchgmapbak.fr ; set cv $fr.grid
    set last [$fr.cs.cs cget -text]
    if { [catch "set p $MapImageFile($dx,$dy)"] } {
	set p ""
	$fr.bns.clr configure -state disabled
    } else {
	if { [string compare $dx,$dy 0,0] } {
	    $fr.bns.clr configure -state normal
	}
    }
    $cv itemconfigure forIm=$last -width 2 -outline black
    $cv itemconfigure forIm=$dx,$dy -width 4 -outline $COLOUR(mapsel)
    .wchgmapbak.fr.cs.cs configure -text $dx,$dy
    .wchgmapbak.fr.pt.pt configure -text $p
    return
}

proc MapBackGridLoad {rw rh} {
    # (re-)load one image for map background
    #  $rw: width of grid rectangle
    #  $rh: height of grid rectangle
    global Map MapImageFile MapImageWidth MapImageHeight MapImageItems \
	    MapImageGrid File MESS TXT COLOUR

    set fr .wchgmapbak.fr
    scan [set cs [$fr.cs.cs cget -text]] %d,%d dx dy
    if { [string compare [set f [GMOpenFile $TXT(loadfrm) Image r]] ".."] } {
	set filename $File(Image)
	if { [BadImage MapImage$cs $filename] } {
	    GMMessage $MESS(badimage)
	    return
	}
	set MapImageFile($cs) [file join [pwd] $filename]
	$fr.pt.pt configure -text $MapImageFile($cs)
	set cv $fr.grid
	$cv itemconfigure forIm=$cs -fill $COLOUR(fullgrid)
	$Map delete forIm=$cs
	set x [expr $MapImageWidth*$dx] ; set y [expr $MapImageHeight*$dy]
	set it [$Map create image $x $y \
		-image "MapImage$cs" -anchor nw \
		-tags [list map mapimage forIm=$cs]]
	$Map lower $it
	lappend MapImageItems $it
	SetMapBounds
	MapWideBackGrid $cv x $dx Column \
		$MapImageGrid(dymin) $MapImageGrid(dyn) $rw $rh
	MapWideBackGrid $cv y $dy Row \
		$MapImageGrid(dxmin) $MapImageGrid(dxn) $rw $rh
	$fr.bns.clr configure -state normal
    }
    return
}

proc MapBackGridClear {} {
    # clear one image from map background
    global Map MapImageFile MapImageItems MapImageGrid COLOUR MESS

    set fr .wchgmapbak.fr
    scan [set cs [$fr.cs.cs cget -text]] %d,%d dx dy
    if { [GMConfirm "$MESS(okclrbkim) $cs"] } {
	set it [$Map find withtag forIm=$cs]
	$Map delete forIm=$cs
	SetMapBounds
	$fr.bns.clr configure -state disabled
	catch "image delete MapImage$cs ; unset MapImageFile($cs)"
	if { [set ix [lsearch -exact $MapImageItems $it]] >= 0 } {
	    set MapImageItems [lreplace $MapImageItems $ix $ix]
	}
	set cv $fr.grid
	$cv itemconfigure forIm=$cs -fill $COLOUR(emptygrid)
	if { ([MapShrinkBackGrid $cv x $dx %d,*] | \
		[MapShrinkBackGrid $cv y $dy *,%d]) && \
		! [string compare [$cv find withtag forIm=$dx,$dy] ""] } {
	    $cv itemconfigure forIm=0,0 -outline $COLOUR(mapsel)
	    $fr.cs.cs configure -text 0,0
	    $fr.pt.pt configure -text $MapImageFile(0,0)
	} else {
	    $fr.pt.pt configure -text ""
	}
    }
    return
}

proc MapColumnBackGrid {gr dx dymin dyn rw rh} {
    # make column of grid in dialog used to change map
    #  background images
    #  $gr: canvas with grid
    #  $dx: grid coordinate
    #  $dymin: min grid y-coordinate
    #  $dyn: number of slots along y-coordinate
    #  $rw, $rh: rectangle dimensions
    global MapImageGrid COLOUR

    set m [expr $dymin+$dyn]
    for { set dy $dymin } { $dy < $m } { incr dy } {
	set it [$gr create rectangle [expr ($dx+1)*$rw+2] \
		[expr ($dy+1)*$rh+2] [expr ($dx+2)*$rw] \
		[expr ($dy+2)*$rh] -width 2 -fill $COLOUR(emptygrid) \
		-tags [list grid forIm=$dx,$dy]]
	$gr bind $it <Enter> "MapBackGridEnter $dx $dy"
	$gr bind $it <Leave> "MapBackGridLeave $dx $dy"
	$gr bind $it <Button-1> "MapBackGridSelect $dx $dy"
    }
    return
}

proc MapRowBackGrid {gr dy dxmin dxn rw rh} {
    # make row of grid in dialog used to change map
    #  background images
    #  $gr: canvas with grid
    #  $dy: grid coordinate
    #  $dxmin: min grid x-coordinate
    #  $dxn: number of slots along x-coordinate
    #  $rw, $rh: rectangle dimensions
    global MapImageGrid COLOUR

    set m [expr $dxmin+$dxn]
    for { set dx $dxmin } { $dx < $m } { incr dx } {
	set it [$gr create rectangle [expr ($dx+1)*$rw+2] \
		[expr ($dy+1)*$rh+2] [expr ($dx+2)*$rw] \
		[expr ($dy+2)*$rh] -width 2 -fill $COLOUR(emptygrid) \
		-tags [list grid forIm=$dx,$dy]]
	$gr bind $it <Enter> "MapBackGridEnter $dx $dy"
	$gr bind $it <Leave> "MapBackGridLeave $dx $dy"
	$gr bind $it <Button-1> "MapBackGridSelect $dx $dy"
    }
    return
}

proc MapWideBackGrid {gr dir c how omin on rw rh} {
    # add external row/column of grid in dialog used to change map
    #  background images if the external row/column becomes non-empty
    #  $gr: canvas with grid
    #  $dir in {x, y}
    #  $c: grid coordinate along $dir
    #  $how in {Row, Column} according to $dir
    #  $omin: min grid coordinate along other direction
    #  $on: number of slots along other direction
    #  $rw, $rh: rectangle dimensions
    global MapImageGrid

    if { $c != 0 } {
	set chg 0 ; set dd d$dir
	if { $c == [set m $MapImageGrid(${dd}min)] } {
	    set chg 1 ; set scr -1
	    incr MapImageGrid(${dd}min) -1 ; incr MapImageGrid(${dd}n)
	    Map${how}BackGrid $gr $MapImageGrid(${dd}min) $omin $on $rw $rh
	} elseif { $c == [expr $MapImageGrid(${dd}n)+$m-1] } {
	    set chg 1 ; set scr 0
	    incr MapImageGrid(${dd}n)
	    Map${how}BackGrid $gr [expr $c+1] $omin $on $rw $rh
	}
	if { $chg } {
	    set bb [$gr bbox all]
	    set x0 [lindex $bb 0] ; set y0 [lindex $bb 1]
	    set x1 [lindex $bb 2] ; set y1 [lindex $bb 3]
	    $gr configure -width [expr $x1-$x0] -height [expr $y1-$y0]
	    $gr ${dir}view scroll $scr units
	}
    }
    return
}

proc MapShrinkBackGrid {gr dir c fmt} {
    # delete external row/column of grid in dialog used to change map
    #  background images if its neighbour becomes empty (external rows
    #  and columns are always empty; minimum size is 3x3, as slot with
    #  origin is never emptied)
    # return 1 if there was shrinking
    #  $gr: canvas with grid
    #  $dir in {x, y}
    #  $c: grid coordinate along $dir
    #  $fmt in {"%d,*", "*,%d"}
    global MapImageGrid MapImageFile

    set chg 0
    if { $c != 0 } {
	set dd d$dir ; set patt [format $fmt $c]
	if { $c == [set c1 [expr $MapImageGrid(${dd}min)+1]] && \
		[NoBackImageAt $patt] } {
	    set chg 1 ; set scr 1 ; incr MapImageGrid(${dd}min)
	} elseif { $c == [expr $MapImageGrid(${dd}n)+$c1-3] && \
		[NoBackImageAt $patt] } {
	    set chg 1 ; set scr 0
	}
	if { $chg } {
	    incr MapImageGrid(${dd}n) -1
            set dc [expr 1-$scr-$scr]
            set cd [expr $c+$dc] ; set patt [format $fmt $cd]
	    foreach it [$gr find withtag grid] {
		if { [lsearch -glob [$gr gettags $it] forIm=$patt] != -1 } {
		    $gr delete $it
		}
	    }
	    set bb [$gr bbox all]
	    set x0 [lindex $bb 0] ; set y0 [lindex $bb 1]
	    set x1 [lindex $bb 2] ; set y1 [lindex $bb 3]
	    $gr configure -width [expr $x1-$x0] -height [expr $y1-$y0]
	    $gr ${dir}view scroll $scr units
	    if { abs($c) != 1 } {
		MapShrinkBackGrid $gr $dir [expr $c-$dc] $fmt
	    }
	}
    }
    return $chg
}

proc NoBackImageAt {patt} {
    # check whether there is a loaded image with coordinates of given pattern
    global MapImageFile

    if { [lsearch -glob [array names MapImageFile] $patt] == -1 } {
	return 1
    }
    return 0
}

proc StatusMapBackMenu {st} {
    # change status of Save/Change/Clear options in Backrgnd menu
    global WConf

    set m $WConf(mapstateback)
    for { set i 1 } { $i < 4 } { incr i } {
	$m entryconfigure $i -state $st
    }
    return
}

#### locate items on map

proc Locate {it} {
    # scroll map to get displayed item on centre
    global Map OVx OVy MAPW2 MAPH2

    if { "[set cs [$Map coords $it]]" != "" } {
	ScrollMapTo [lindex $cs 0] [lindex $cs 1] \
		[expr $OVx+$MAPW2] [expr $OVy+$MAPH2]
    }
    return
}

proc FillLocateMenu {menu} {
    # fill menu with items currently displayed on map
    global Map TXT MAXMENUITEMS RTIdNumber RTWPoints TRName

    if { [winfo exists $menu] } {
	destroy $menu
    }
    set its "" ; set ls ""
    set rts ""
    set tn $TXT(WP)
    foreach it [$Map find withtag WP] {
	set ts [$Map gettags $it]
	if { [set k [lsearch $ts WP=*]] != -1 } {
	    regsub WP= [lindex $ts $k] "" n
	    lappend its $it ; lappend ls "$tn: $n"
	}
	if { [set k [lsearch -glob $ts {inRT=[0-9]*}]] != -1 } {
	    regsub inRT= [lindex $ts $k] "" ix
	    if { [lsearch $rts $ix] == -1 } { lappend rts $ix }
	}
    }
    set tn $TXT(RT)
    foreach ix $rts {
	set fwpn [lindex $RTWPoints($ix) 0]
	if { "[set it [$Map find withtag WP=$fwpn]]" != "" } {
	    lappend its $it ; lappend ls "$tn: $RTIdNumber($ix)"
	}
    }
    set tn $TXT(TR)
    foreach it [$Map find withtag TRfirst] {
	set ts [$Map gettags $it]
	if { [set k [lsearch $ts TR=*]] != -1 } {
	    regsub TR= [lindex $ts $k] "" ix
	    if { $ix != -1 } {
		lappend its $it ; lappend ls "$tn: $TRName($ix)"
	    }
	}
    }
    set mw $menu ; menu $mw -tearoff 0
    set n 0 ; set m 0
    foreach it $its lb $ls {
	if { $n > $MAXMENUITEMS } {
	    $mw add cascade -label "$TXT(more) ..." -menu $mw.m$m
	    set mw $mw.m$m ; menu $mw -tearoff 0
	    set n 0 ; incr m
	}
	$mw add command -label $lb -command "Locate $it"
	incr n
    }
    return
}
