#!/bin/sh
# This is a Tcl/Tk script to be interpreted by wish (Tk8.0 or better): \
exec wish "$0" "$@"

#
#  gpsman --- GPS Manager: a manager for GPS receiver data
#
#  Copyright (c) 2000 Miguel Filgueiras (mig@ncc.up.pt) / Universidade do Porto
#
#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.
#
## Incorporates contributions by Brian Baulch (baulchb@onthenet.com.au)
#      Copyright (c) 1999-2000 by Brian Baulch
#      that allow for the support of Lowrance receivers
##
#
## German language support by Andreas Lange (andreas.lange@rhein-main.de)
#      Copyright (c) 2000 by Andreas Lange
##
#
## Italian language support by Alessandro Palmas (alpalmas@tin.it)
#      Copyright (c) 2000 by Alessandro Palmas
##
#
## Includes a translation of a Perl script by Niki Hammler
#      (http://www.nobaq.net)
#      that converts exported FUGAWI data to DDD GPSman data
##
#
## Some algorithms, formulae and geodetic information taken or adopted from
#  gpstrans - a program to communicate with garmin gps
#      containing parts taken from John F. Waers (jfwaers@csn.net)
#      program MacGPS.
#        Copyright (c) 1995 by Carsten Tschach (tschach@zedat.fu-berlin.de)
#      Uniform Finish grid support:
#        Copyright (c) 1996 Janne Sinkkonen (janne@iki.fi) 
#  INVERSE, FORWARD, INVERS3D, FORWRD3D - programs to compute distances and
#      azimuths between two points.
#      Available from ftp://www.ngs.noaa.gov/pub/pcsoft/for_inv.3d/
#  PROJ4.0 - cartographic projection software
#      by Gerald I. Evenden (gie@kai.er.usgs.gov)
#  International Institute for Aerospace Survey and Earth Sciences (ITC),
#      Enschede, http://kartoweb.itc.nl/geometrics
#
# look for this notice in the corresponding files
##
#
## Some formulae and geodetic information kindly provided by
#      Luisa Bastos, Universidade do Porto
#      Gil Goncalves, Universidade de Coimbra
#      Jose Alberto Goncalves, Universidade do Porto
#      Sergio Cunha, Universidade do Porto
#      Peter H. Dana, University of Texas
##
#
## Correction of bugs in the conversion of UTM/UPS coordinates provided by
#      Peter H. Dana, University of Texas
##
#
#  File: gpsman.tcl
#  Last change:  2 November 2000
#

############ configuration parameters
#
# the first 3 variables below MUST be correctly set for GPSMan to work!
#

 # NOTE for non-Unix users: use "/" (not "\") in pathnames

 # path to serial port; make sure all users have read/write permission

switch $tcl_platform(platform) {
    unix {
	# on Unix systems the path is passed as an argument
	set SERIALPORT [lindex $argv 0]
    }
    default {
	# on Apple Macintoshes the serial port cannot be used under Tcl/Tk
	# on MS-Windows it could be as follows
	set SERIALPORT com2:
    }
}

 # path to directory containing program files and to default preferences file

set SRCDIR gmsrc
    # options file: default path
set OPTFILE ~/.gpsman

## the following values can be set from the options dialog of the interface

 # language: see lang*.tcl files, ALLMONTH array below, main.tcl KNOWNLANGS
    # currently available: engl for English, port for Portuguese, deutsch for
    #   German, it for Italian

set LANG engl

 # ISO latin1 character composition (accents, cedilla, ...)
set ISOLATIN1 1

 # Delete-key to delete last character
set DELETE 1

 # Configuration of main window in {map, lists, single}
set MWINDOWSCONF map

 # GPS receiver trade mark and model (if applicable): give trade mark
 #   (one of: Garmin, Lowrance) followed by model name.
 #   See file recmodels.tcl for list of known models

set GPSREC "Garmin"

 # GPS receiver dependent values (set to Garmin GPS II)

set NAMELENGTH 6
set COMMENTLENGTH 16
set MAXWPOINTS 250
set MAXROUTES 20
set MAXWPINROUTE 30
set MAXTPOINTS 768
set CREATIONDATE 0
set NOLOWERCASE 1

    # sampling interval for Lowrance receivers in seconds
set INTERVAL 3

   # default symbol name (GPSMan convention; see symbols.tcl)
set DEFAULTSYMBOL WP_dot
    #  and symbol display options (GPSMan convention; see symbols.tcl)
set DEFAULTDISPOPT s_name

 # data items with same name will be overwritten (ovwrt) or created under
 #  new name (rename):
set EQNAMEDATA ovwrt

 # keep hidden data (not available for editing) when changing item;
 #  one of: never, ask (at every change), always
set KEEPHIDDEN never

 # default values

         # possible distance units: KM NAUTMILE STATMILE
set DISTUNIT KM

         # possible position formats: DMS (hdeg, min, sec), DMM (hdeg, min),
         #                            DDD (hdeg), UTM/UPS (ze, zn, e, n),
         #                            BNG (British grid: zone, e, n),
         #                            CMP (Portuguese military grid: e, n),
         #                            GKK (German grid: zone, e, n),
         #                            ITM (Irish grid: zone, e, n)
set PositionFormat DMS
set Datum "WGS 84"
set TimeOffset 1

         # possible date formats: YYYYMMDD  (1998.03.20 09:35:23),
         #                        MMDDYYYY  (03/20/1998 09:35:23),
         #                        DDMMMYYYY (20-Mar-1998 09:35:23)
set DateFormat DDMMMYYYY

  # formulae: either more accurate or faster
set ACCFORMULAE 0

  # ask for confirmation of projection parameters
set ASKPROJPARAMS 1

 # geometry

  # map dimensions in pixels
set MAPWIDTH 800 ; set MAPHEIGHT 550
         # length of scale arrow (pixels)
set MAPSCLENGTH 300
         # initial map scale: 1, 2, 4, 8, ... 512*$DISTUNIT/$MAPSCLENGTH pixel
         #  or, if $DISTUNIT==km: 50, 100, 200, 500 m/$MAPSCLENGTH pixel
         #      else: 150, 300, 600, 2400 ft/$MAPSCLENGTH pixel
         # actual distance and unit (from the values above) for the scale arrow
set MAPSCALE "4 km"

  # menus
set MAXMENUITEMS 20

  # dimensions of lists
set LISTWIDTH 9
set LISTHEIGHT 22

  # initial positions
     # single main window
set IPOSX 5 ; set IPOSY 40
     # lists window
set LPOSX 40 ; set LPOSY 25
     # map window
set MPOSX 5 ; set MPOSY 40
     # receiver window
set RPOSX 10 ; set RPOSY 490
     # errors, messages, warnings
set EPOSX 340 ; set EPOSY 50
     # dialogs
set DPOSX 290 ; set DPOSY 50
     # offset for dialog for making route on map
set DPOSRTMAP -200

  # colours
       # general fore/background
set COLOUR(fg) "#090980"
set COLOUR(bg) "#f7f7e0"
       # error background
set COLOUR(messbg) "#ff8d90"
       # confirmation background
set COLOUR(confbg) "#F0E9C0"
       # selection background
set COLOUR(selbg) "#F0E9C0"
       # input dialog background
set COLOUR(dialbg) gray
       # receiver offline
set COLOUR(offline) "#d83232"
       # receiver online
set COLOUR(online) green
       # selected checkbutton
set COLOUR(check) red
       # selected map item
set COLOUR(mapsel) red
       # wayponts on map
set COLOUR(WP) "#ff6400"
       # routes on map
set COLOUR(RT) "#48C1BC"
       # route being defined on map
set COLOUR(mkRT) "#469e36"
       # tracks on map
set COLOUR(TR) "#EF9A00"
       # track points on map
set COLOUR(TP) "#CD69C9"
       # map legends
set COLOUR(mapleg) "#8B0000"
       # lines in map animation
set COLOUR(anim) darkblue
       # slots in map image grid
set COLOUR(emptygrid) white
set COLOUR(selemptygrid) "#ffe080"
set COLOUR(fullgrid) "#00d8d8"
set COLOUR(selfullgrid) "#ffd8d8"

 # permissions of created files
set PERMS 0640

 # default paper size (one of those whose dimensions are given below)

set PAPERSIZE a4

### the following values can be configured if needs be

  # paper size without margins

set PAGEWIDTH(a4) 184m ; set PAGEHEIGHT(a4) 271m
set PAGEWIDTH(a5) 134m ; set PAGEHEIGHT(a5) 195m
set PAGEWIDTH(b4) 250m ; set PAGEHEIGHT(b4) 353m
set PAGEWIDTH(b5) 176m ; set PAGEHEIGHT(b5) 250m

  # possible abbreviations for months in all known languages
  # English abbreviations must not be deleted
set ALLMONTH(1) "Jan Gen"
set ALLMONTH(2) "Feb Fev"
set ALLMONTH(3) "Mar Mr"
set ALLMONTH(4) "Apr Abr"
set ALLMONTH(5) "May Mai Mag"
set ALLMONTH(6) "Jun Giu"
set ALLMONTH(7) "Jul Lug"
set ALLMONTH(8) "Aug Ago"
set ALLMONTH(9) "Sep Set"
set ALLMONTH(10) "Oct Okt Out Ott"
set ALLMONTH(11) "Nov"
set ALLMONTH(12) "Dec Dez Dic"


###### NO CONFIGURABLE VALUES AFTER THIS LINE ##############################

set InitOpts 0

if { [file readable [set fname [file tail $OPTFILE]]] } {
    source $fname
    set OPTFILE [file join [pwd] $fname]
} elseif { [file readable $OPTFILE] } {
    source $OPTFILE
} else {
    set InitOpts 1
}


# source files read before initialization of preferences file
set SRCFILESBEGIN "lang$LANG symbols gendials util check \
           geod compute posncomp recmodels options"

# source files setting receiver parameters
set SRCFILESREC "recdefs"

# source files depending on receiver trade mark
set SRCFILESFOR(Garmin) "garmin_protocols garmin_symbols garmin"
set SRCFILESFOR(Lowrance) "lowrance_symbols lowrance"

# conditional source files
set SRCFILESCOND [list "ACCFORMULAE acccomp.tcl"]
# other generic source files
set SRCFILESEND "wrtdials search lists files gpsinfo mapproj maptransf map \
	mapanim know trtort"

source $SRCDIR/main.tcl

GMStart
GMInit

